# -*- coding: utf-8 -*-
"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""
from __future__ import absolute_import, print_function, unicode_literals

import copy
import datetime
import logging

import salt.config
import salt.utils.schedule
from salt.utils.schedule import Schedule
from tests.support.helpers import slowTest
from tests.support.mock import MagicMock, patch
from tests.support.unit import skipIf
from tests.unit.utils.scheduler.base import SchedulerTestsBase

# pylint: disable=import-error,unused-import
try:
    import croniter

    _CRON_SUPPORTED = True
except ImportError:
    _CRON_SUPPORTED = False
# pylint: enable=import-error

log = logging.getLogger(__name__)


# pylint: disable=too-many-public-methods,invalid-name
class ScheduleTestCase(SchedulerTestsBase):
    """
    Unit tests for salt.utils.schedule module
    """

    # delete_job tests

    @slowTest
    def test_delete_job_exists(self):
        """
        Tests ensuring the job exists and deleting it
        """
        self.schedule.opts.update({"schedule": {"foo": "bar"}, "pillar": {}})
        self.assertIn("foo", self.schedule.opts["schedule"])
        self.schedule.delete_job("foo")
        self.assertNotIn("foo", self.schedule.opts["schedule"])

    @slowTest
    def test_delete_job_in_pillar(self):
        """
        Tests ignoring deletion job from pillar
        """
        self.schedule.opts.update(
            {"pillar": {"schedule": {"foo": "bar"}}, "schedule": {}}
        )
        self.assertIn("foo", self.schedule.opts["pillar"]["schedule"])
        self.schedule.delete_job("foo")
        self.assertIn("foo", self.schedule.opts["pillar"]["schedule"])

    @slowTest
    def test_delete_job_intervals(self):
        """
        Tests removing job from intervals
        """
        self.schedule.opts.update({"pillar": {}, "schedule": {}})
        self.schedule.intervals = {"foo": "bar"}
        self.schedule.delete_job("foo")
        self.assertNotIn("foo", self.schedule.intervals)

    @slowTest
    def test_delete_job_prefix(self):
        """
        Tests ensuring jobs exists and deleting them by prefix
        """
        self.schedule.opts.update(
            {
                "schedule": {"foobar": "bar", "foobaz": "baz", "fooboo": "boo"},
                "pillar": {},
            }
        )
        ret = copy.deepcopy(self.schedule.opts)
        del ret["schedule"]["foobar"]
        del ret["schedule"]["foobaz"]
        self.schedule.delete_job_prefix("fooba")
        self.assertEqual(self.schedule.opts, ret)

    @slowTest
    def test_delete_job_prefix_in_pillar(self):
        """
        Tests ignoring deletion jobs by prefix from pillar
        """
        self.schedule.opts.update(
            {
                "pillar": {
                    "schedule": {"foobar": "bar", "foobaz": "baz", "fooboo": "boo"}
                },
                "schedule": {},
            }
        )
        ret = copy.deepcopy(self.schedule.opts)
        self.schedule.delete_job_prefix("fooba")
        self.assertEqual(self.schedule.opts, ret)

    # add_job tests

    def test_add_job_data_not_dict(self):
        """
        Tests if data is a dictionary
        """
        data = "foo"
        self.assertRaises(ValueError, Schedule.add_job, self.schedule, data)

    def test_add_job_multiple_jobs(self):
        """
        Tests if more than one job is scheduled at a time
        """
        data = {"key1": "value1", "key2": "value2"}
        self.assertRaises(ValueError, Schedule.add_job, self.schedule, data)

    @slowTest
    def test_add_job(self):
        """
        Tests adding a job to the schedule
        """
        data = {"foo": {"bar": "baz"}}
        ret = copy.deepcopy(self.schedule.opts)
        ret.update(
            {
                "schedule": {
                    "foo": {"bar": "baz", "enabled": True},
                    "hello": {"world": "peace", "enabled": True},
                },
                "pillar": {},
            }
        )
        self.schedule.opts.update(
            {"schedule": {"hello": {"world": "peace", "enabled": True}}, "pillar": {}}
        )
        Schedule.add_job(self.schedule, data)
        self.assertEqual(self.schedule.opts, ret)

    # enable_job tests

    @slowTest
    def test_enable_job(self):
        """
        Tests enabling a job
        """
        self.schedule.opts.update({"schedule": {"name": {"enabled": "foo"}}})
        Schedule.enable_job(self.schedule, "name")
        self.assertTrue(self.schedule.opts["schedule"]["name"]["enabled"])

    @slowTest
    def test_enable_job_pillar(self):
        """
        Tests ignoring enable a job from pillar
        """
        self.schedule.opts.update(
            {"pillar": {"schedule": {"name": {"enabled": False}}}}
        )
        Schedule.enable_job(self.schedule, "name", persist=False)
        self.assertFalse(self.schedule.opts["pillar"]["schedule"]["name"]["enabled"])

    # disable_job tests

    @slowTest
    def test_disable_job(self):
        """
        Tests disabling a job
        """
        self.schedule.opts.update(
            {"schedule": {"name": {"enabled": "foo"}}, "pillar": {}}
        )
        Schedule.disable_job(self.schedule, "name")
        self.assertFalse(self.schedule.opts["schedule"]["name"]["enabled"])

    @slowTest
    def test_disable_job_pillar(self):
        """
        Tests ignoring disable a job in pillar
        """
        self.schedule.opts.update(
            {"pillar": {"schedule": {"name": {"enabled": True}}}, "schedule": {}}
        )
        Schedule.disable_job(self.schedule, "name", persist=False)
        self.assertTrue(self.schedule.opts["pillar"]["schedule"]["name"]["enabled"])

    # modify_job tests

    @slowTest
    def test_modify_job(self):
        """
        Tests modifying a job in the scheduler
        """
        schedule = {"foo": "bar"}
        self.schedule.opts.update({"schedule": {"name": "baz"}, "pillar": {}})
        ret = copy.deepcopy(self.schedule.opts)
        ret.update({"schedule": {"name": {"foo": "bar"}}})
        Schedule.modify_job(self.schedule, "name", schedule)
        self.assertEqual(self.schedule.opts, ret)

    def test_modify_job_not_exists(self):
        """
        Tests modifying a job in the scheduler if jobs not exists
        """
        schedule = {"foo": "bar"}
        self.schedule.opts.update({"schedule": {}, "pillar": {}})
        ret = copy.deepcopy(self.schedule.opts)
        ret.update({"schedule": {"name": {"foo": "bar"}}})
        Schedule.modify_job(self.schedule, "name", schedule)
        self.assertEqual(self.schedule.opts, ret)

    def test_modify_job_pillar(self):
        """
        Tests ignoring modification of job from pillar
        """
        schedule = {"foo": "bar"}
        self.schedule.opts.update(
            {"schedule": {}, "pillar": {"schedule": {"name": "baz"}}}
        )
        ret = copy.deepcopy(self.schedule.opts)
        Schedule.modify_job(self.schedule, "name", schedule, persist=False)
        self.assertEqual(self.schedule.opts, ret)

    maxDiff = None

    # enable_schedule tests

    @slowTest
    def test_enable_schedule(self):
        """
        Tests enabling the scheduler
        """
        with patch(
            "salt.utils.schedule.Schedule.persist", MagicMock(return_value=None)
        ) as persist_mock:
            self.schedule.opts.update({"schedule": {"enabled": "foo"}, "pillar": {}})
            Schedule.enable_schedule(self.schedule)
            self.assertTrue(self.schedule.opts["schedule"]["enabled"])

        persist_mock.assert_called()

    # disable_schedule tests

    @slowTest
    def test_disable_schedule(self):
        """
        Tests disabling the scheduler
        """
        with patch(
            "salt.utils.schedule.Schedule.persist", MagicMock(return_value=None)
        ) as persist_mock:
            self.schedule.opts.update({"schedule": {"enabled": "foo"}, "pillar": {}})
            Schedule.disable_schedule(self.schedule)
            self.assertFalse(self.schedule.opts["schedule"]["enabled"])

        persist_mock.assert_called()

    # reload tests

    def test_reload_update_schedule_key(self):
        """
        Tests reloading the schedule from saved schedule where both the
        saved schedule and self.schedule.opts contain a schedule key
        """
        saved = {"schedule": {"foo": "bar"}}
        ret = copy.deepcopy(self.schedule.opts)
        ret.update({"schedule": {"foo": "bar", "hello": "world"}})
        self.schedule.opts.update({"schedule": {"hello": "world"}})
        Schedule.reload(self.schedule, saved)
        self.assertEqual(self.schedule.opts, ret)

    def test_reload_update_schedule_no_key(self):
        """
        Tests reloading the schedule from saved schedule that does not
        contain a schedule key but self.schedule.opts does
        """
        saved = {"foo": "bar"}
        ret = copy.deepcopy(self.schedule.opts)
        ret.update({"schedule": {"foo": "bar", "hello": "world"}})
        self.schedule.opts.update({"schedule": {"hello": "world"}})
        Schedule.reload(self.schedule, saved)
        self.assertEqual(self.schedule.opts, ret)

    def test_reload_no_schedule_in_opts(self):
        """
        Tests reloading the schedule from saved schedule that does not
        contain a schedule key and neither does self.schedule.opts
        """
        saved = {"foo": "bar"}
        ret = copy.deepcopy(self.schedule.opts)
        ret["schedule"] = {"foo": "bar"}
        self.schedule.opts.pop("schedule", None)
        Schedule.reload(self.schedule, saved)
        self.assertEqual(self.schedule.opts, ret)

    def test_reload_schedule_in_saved_but_not_opts(self):
        """
        Tests reloading the schedule from saved schedule that contains
        a schedule key, but self.schedule.opts does not
        """
        saved = {"schedule": {"foo": "bar"}}
        ret = copy.deepcopy(self.schedule.opts)
        ret["schedule"] = {"foo": "bar"}
        self.schedule.opts.pop("schedule", None)
        Schedule.reload(self.schedule, saved)
        self.assertEqual(self.schedule.opts, ret)

    # eval tests

    def test_eval_schedule_is_not_dict(self):
        """
        Tests eval if the schedule is not a dictionary
        """
        self.schedule.opts.update({"schedule": "", "pillar": {"schedule": {}}})
        self.assertRaises(ValueError, Schedule.eval, self.schedule)

    def test_eval_schedule_is_not_dict_in_pillar(self):
        """
        Tests eval if the schedule from pillar is not a dictionary
        """
        self.schedule.opts.update({"schedule": {}, "pillar": {"schedule": ""}})
        self.assertRaises(ValueError, Schedule.eval, self.schedule)

    def test_eval_schedule_time(self):
        """
        Tests eval if the schedule setting time is in the future
        """
        self.schedule.opts.update({"pillar": {"schedule": {}}})
        self.schedule.opts.update(
            {"schedule": {"testjob": {"function": "test.true", "seconds": 60}}}
        )
        now = datetime.datetime.now()
        self.schedule.eval()
        self.assertTrue(
            self.schedule.opts["schedule"]["testjob"]["_next_fire_time"] > now
        )

    def test_eval_schedule_time_eval(self):
        """
        Tests eval if the schedule setting time is in the future plus splay
        """
        self.schedule.opts.update({"pillar": {"schedule": {}}})
        self.schedule.opts.update(
            {
                "schedule": {
                    "testjob": {"function": "test.true", "seconds": 60, "splay": 5}
                }
            }
        )
        now = datetime.datetime.now()
        self.schedule.eval()
        self.assertTrue(
            self.schedule.opts["schedule"]["testjob"]["_splay"] - now
            > datetime.timedelta(seconds=60)
        )

    @skipIf(not _CRON_SUPPORTED, "croniter module not installed")
    def test_eval_schedule_cron(self):
        """
        Tests eval if the schedule is defined with cron expression
        """
        self.schedule.opts.update({"pillar": {"schedule": {}}})
        self.schedule.opts.update(
            {"schedule": {"testjob": {"function": "test.true", "cron": "* * * * *"}}}
        )
        now = datetime.datetime.now()
        self.schedule.eval()
        self.assertTrue(
            self.schedule.opts["schedule"]["testjob"]["_next_fire_time"] > now
        )

    @skipIf(not _CRON_SUPPORTED, "croniter module not installed")
    def test_eval_schedule_cron_splay(self):
        """
        Tests eval if the schedule is defined with cron expression plus splay
        """
        self.schedule.opts.update({"pillar": {"schedule": {}}})
        self.schedule.opts.update(
            {
                "schedule": {
                    "testjob": {
                        "function": "test.true",
                        "cron": "* * * * *",
                        "splay": 5,
                    }
                }
            }
        )
        self.schedule.eval()
        self.assertTrue(
            self.schedule.opts["schedule"]["testjob"]["_splay"]
            > self.schedule.opts["schedule"]["testjob"]["_next_fire_time"]
        )

    @slowTest
    def test_handle_func_schedule_minion_blackout(self):
        """
        Tests eval if the schedule from pillar is not a dictionary
        """
        self.schedule.opts.update({"pillar": {"schedule": {}}})
        self.schedule.opts.update({"grains": {"minion_blackout": True}})

        self.schedule.opts.update(
            {"schedule": {"testjob": {"function": "test.true", "seconds": 60}}}
        )
        data = {
            "function": "test.true",
            "_next_scheduled_fire_time": datetime.datetime(
                2018, 11, 21, 14, 9, 53, 903438
            ),
            "run": True,
            "name": "testjob",
            "seconds": 60,
            "_splay": None,
            "_seconds": 60,
            "jid_include": True,
            "maxrunning": 1,
            "_next_fire_time": datetime.datetime(2018, 11, 21, 14, 8, 53, 903438),
        }

        with patch.object(salt.utils.schedule, "log") as log_mock:
            with patch("salt.utils.process.daemonize"), patch("sys.platform", "linux2"):
                self.schedule.handle_func(False, "test.ping", data)
                self.assertTrue(log_mock.exception.called)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import shutil
import sys
import tempfile
import textwrap
import threading
import time

import pytest
import salt.utils.atomicfile
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt.ext import six
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES
from tests.support.case import ModuleCase
from tests.support.helpers import slowTest, with_tempdir
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.sminion import create_sminion
from tests.support.unit import skipIf

log = logging.getLogger(__name__)


DEFAULT_ENDING = salt.utils.stringutils.to_bytes(os.linesep)


@pytest.mark.windows_whitelisted
class StateModuleTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the state module
    """

    maxDiff = None

    @classmethod
    def setUpClass(cls):
        def _reline(path, ending=DEFAULT_ENDING):
            """
            Normalize the line endings of a file.
            """
            with salt.utils.files.fopen(path, "rb") as fhr:
                lines = fhr.read().splitlines()
            with salt.utils.atomicfile.atomic_open(path, "wb") as fhw:
                for line in lines:
                    fhw.write(line + ending)

        destpath = os.path.join(RUNTIME_VARS.BASE_FILES, "testappend", "firstif")
        _reline(destpath)
        destpath = os.path.join(RUNTIME_VARS.BASE_FILES, "testappend", "secondif")
        _reline(destpath)
        if salt.utils.platform.is_windows():
            cls.TIMEOUT = 600
            # Be sure to have everything sync'ed
            sminion = create_sminion()
            sminion.functions.saltutil.sync_all()
        else:
            cls.TIMEOUT = 10

    @slowTest
    def test_show_highstate(self):
        """
        state.show_highstate
        """
        high = self.run_function("state.show_highstate")
        destpath = os.path.join(RUNTIME_VARS.TMP, "testfile")
        self.assertTrue(isinstance(high, dict))
        self.assertTrue(destpath in high)
        self.assertEqual(high[destpath]["__env__"], "base")

    @slowTest
    def test_show_lowstate(self):
        """
        state.show_lowstate
        """
        low = self.run_function("state.show_lowstate")
        self.assertTrue(isinstance(low, list))
        self.assertTrue(isinstance(low[0], dict))

    @slowTest
    def test_show_states(self):
        """
        state.show_states
        """
        states = self.run_function("state.show_states")
        self.assertTrue(isinstance(states, list))
        self.assertTrue(isinstance(states[0], six.string_types))

        states = self.run_function("state.show_states", sorted=False)
        self.assertTrue(isinstance(states, list))
        self.assertTrue(isinstance(states[0], six.string_types))

    @slowTest
    def test_show_states_missing_sls(self):
        """
        Test state.show_states with a sls file
        defined in a top file is missing
        """
        topfile = os.path.join(RUNTIME_VARS.TMP_STATE_TREE, "top.sls")
        with salt.utils.files.fopen(topfile, "w") as top_file:
            top_file.write(
                textwrap.dedent(
                    """\
                                           base:
                                             '*':
                                               - doesnotexist
                                           """
                )
            )
        states = self.run_function("state.show_states")
        assert isinstance(states, list)
        assert states == ["No matching sls found for 'doesnotexist' in env 'base'"]

    @slowTest
    def test_catch_recurse(self):
        """
        state.show_sls used to catch a recursive ref
        """
        err = self.run_function("state.sls", mods="recurse_fail")
        self.assertIn("recursive", err[0])

    @slowTest
    def test_no_recurse(self):
        """
        verify that a sls structure is NOT a recursive ref
        """
        sls = self.run_function("state.show_sls", mods="recurse_ok")
        self.assertIn("snmpd", sls)

    @slowTest
    def test_no_recurse_two(self):
        """
        verify that a sls structure is NOT a recursive ref
        """
        sls = self.run_function("state.show_sls", mods="recurse_ok_two")
        self.assertIn("/etc/nagios/nrpe.cfg", sls)

    @slowTest
    def test_running_dictionary_consistency(self):
        """
        Test the structure of the running dictionary so we don't change it
        without deprecating/documenting the change
        """
        running_dict_fields = [
            "__id__",
            "__run_num__",
            "__sls__",
            "changes",
            "comment",
            "duration",
            "name",
            "result",
            "start_time",
        ]

        sls = self.run_function(
            "state.single", fun="test.succeed_with_changes", name="gndn"
        )

        for state, ret in sls.items():
            for field in running_dict_fields:
                self.assertIn(field, ret)

    @slowTest
    def test_running_dictionary_key_sls(self):
        """
        Ensure the __sls__ key is either null or a string
        """
        sls1 = self.run_function(
            "state.single", fun="test.succeed_with_changes", name="gndn"
        )

        sls2 = self.run_function("state.sls", mods="gndn")

        for state, ret in sls1.items():
            self.assertTrue(isinstance(ret["__sls__"], type(None)))

        for state, ret in sls2.items():
            self.assertTrue(isinstance(ret["__sls__"], six.string_types))

    def _remove_request_cache_file(self):
        """
        remove minion state request file
        """
        cache_file = os.path.join(self.get_config("minion")["cachedir"], "req_state.p")
        if os.path.exists(cache_file):
            os.remove(cache_file)

    @slowTest
    def test_request(self):
        """
        verify sending a state request to the minion(s)
        """
        self._remove_request_cache_file()

        ret = self.run_function("state.request", mods="modules.state.requested")
        result = ret["cmd_|-count_root_dir_contents_|-ls -a / | wc -l_|-run"]["result"]
        self.assertEqual(result, None)

    @slowTest
    def test_check_request(self):
        """
        verify checking a state request sent to the minion(s)
        """
        self._remove_request_cache_file()

        self.run_function("state.request", mods="modules.state.requested")
        ret = self.run_function("state.check_request")
        result = ret["default"]["test_run"][
            "cmd_|-count_root_dir_contents_|-ls -a / | wc -l_|-run"
        ]["result"]
        self.assertEqual(result, None)

    @slowTest
    def test_clear_request(self):
        """
        verify clearing a state request sent to the minion(s)
        """
        self._remove_request_cache_file()

        self.run_function("state.request", mods="modules.state.requested")
        ret = self.run_function("state.clear_request")
        self.assertTrue(ret)

    @slowTest
    def test_run_request_succeeded(self):
        """
        verify running a state request sent to the minion(s)
        """
        self._remove_request_cache_file()

        if salt.utils.platform.is_windows():
            self.run_function("state.request", mods="modules.state.requested_win")
        else:
            self.run_function("state.request", mods="modules.state.requested")

        ret = self.run_function("state.run_request")

        if salt.utils.platform.is_windows():
            key = "cmd_|-count_root_dir_contents_|-Get-ChildItem C:\\\\ | Measure-Object | %{$_.Count}_|-run"
        else:
            key = "cmd_|-count_root_dir_contents_|-ls -a / | wc -l_|-run"

        result = ret[key]["result"]
        self.assertTrue(result)

    @slowTest
    def test_run_request_failed_no_request_staged(self):
        """
        verify not running a state request sent to the minion(s)
        """
        self._remove_request_cache_file()

        self.run_function("state.request", mods="modules.state.requested")
        self.run_function("state.clear_request")
        ret = self.run_function("state.run_request")
        self.assertEqual(ret, {})

    @with_tempdir()
    @slowTest
    def test_issue_1896_file_append_source(self, base_dir):
        """
        Verify that we can append a file's contents
        """
        testfile = os.path.join(base_dir, "test.append")

        ret = self.run_state("file.touch", name=testfile)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "file.append", name=testfile, source="salt://testappend/firstif"
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "file.append", name=testfile, source="salt://testappend/secondif"
        )
        self.assertSaltTrueReturn(ret)

        with salt.utils.files.fopen(testfile, "r") as fp_:
            testfile_contents = salt.utils.stringutils.to_unicode(fp_.read())

        contents = textwrap.dedent(
            """\
            # set variable identifying the chroot you work in (used in the prompt below)
            if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
                debian_chroot=$(cat /etc/debian_chroot)
            fi

            # enable bash completion in interactive shells
            if [ -f /etc/bash_completion ] && ! shopt -oq posix; then
                . /etc/bash_completion
            fi
            """
        )

        if salt.utils.platform.is_windows():
            new_contents = contents.splitlines()
            contents = os.linesep.join(new_contents)
            contents += os.linesep

        self.assertMultiLineEqual(contents, testfile_contents)

        ret = self.run_state(
            "file.append", name=testfile, source="salt://testappend/secondif"
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "file.append", name=testfile, source="salt://testappend/firstif"
        )
        self.assertSaltTrueReturn(ret)

        with salt.utils.files.fopen(testfile, "r") as fp_:
            testfile_contents = salt.utils.stringutils.to_unicode(fp_.read())

        self.assertMultiLineEqual(contents, testfile_contents)

    @slowTest
    def test_issue_1876_syntax_error(self):
        """
        verify that we catch the following syntax error::

            /tmp/salttest/issue-1876:

              file:
                - managed
                - source: salt://testfile

              file.append:
                - text: foo

        """
        testfile = os.path.join(RUNTIME_VARS.TMP, "issue-1876")

        sls = self.run_function("state.sls", mods="issue-1876")
        self.assertIn(
            "ID '{0}' in SLS 'issue-1876' contains multiple state "
            "declarations of the same type".format(testfile),
            sls,
        )

    @slowTest
    def test_issue_1879_too_simple_contains_check(self):
        expected = textwrap.dedent(
            """\
            # set variable identifying the chroot you work in (used in the prompt below)
            if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
                debian_chroot=$(cat /etc/debian_chroot)
            fi
            # enable bash completion in interactive shells
            if [ -f /etc/bash_completion ] && ! shopt -oq posix; then
                . /etc/bash_completion
            fi
            """
        )

        if salt.utils.platform.is_windows():
            new_contents = expected.splitlines()
            expected = os.linesep.join(new_contents)
            expected += os.linesep

        testfile = os.path.join(RUNTIME_VARS.TMP, "issue-1879")
        # Delete if exiting
        if os.path.isfile(testfile):
            os.unlink(testfile)

        # Create the file
        ret = self.run_function("state.sls", mods="issue-1879", timeout=120)
        self.assertSaltTrueReturn(ret)

        # The first append
        ret = self.run_function("state.sls", mods="issue-1879.step-1", timeout=120)
        self.assertSaltTrueReturn(ret)

        # The second append
        ret = self.run_function("state.sls", mods="issue-1879.step-2", timeout=120)
        self.assertSaltTrueReturn(ret)

        # Does it match?
        try:
            with salt.utils.files.fopen(testfile, "r") as fp_:
                contents = salt.utils.stringutils.to_unicode(fp_.read())
            self.assertMultiLineEqual(expected, contents)
            # Make sure we don't re-append existing text
            ret = self.run_function("state.sls", mods="issue-1879.step-1", timeout=120)
            self.assertSaltTrueReturn(ret)

            ret = self.run_function("state.sls", mods="issue-1879.step-2", timeout=120)
            self.assertSaltTrueReturn(ret)

            with salt.utils.files.fopen(testfile, "r") as fp_:
                contents = salt.utils.stringutils.to_unicode(fp_.read())
            self.assertMultiLineEqual(expected, contents)
        except Exception:  # pylint: disable=broad-except
            if os.path.exists(testfile):
                shutil.copy(testfile, testfile + ".bak")
            raise
        finally:
            if os.path.exists(testfile):
                os.unlink(testfile)

    @slowTest
    def test_include(self):
        tempdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, tempdir, ignore_errors=True)
        pillar = {}
        for path in ("include-test", "to-include-test", "exclude-test"):
            pillar[path] = os.path.join(tempdir, path)
        ret = self.run_function("state.sls", mods="include-test", pillar=pillar)
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isfile(pillar["include-test"]))
        self.assertTrue(os.path.isfile(pillar["to-include-test"]))
        self.assertFalse(os.path.isfile(pillar["exclude-test"]))

    @slowTest
    def test_exclude(self):
        tempdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, tempdir, ignore_errors=True)
        pillar = {}
        for path in ("include-test", "exclude-test", "to-include-test"):
            pillar[path] = os.path.join(tempdir, path)
        ret = self.run_function("state.sls", mods="exclude-test", pillar=pillar)
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isfile(pillar["include-test"]))
        self.assertTrue(os.path.isfile(pillar["exclude-test"]))
        self.assertFalse(os.path.isfile(pillar["to-include-test"]))

    @skipIf(
        salt.utils.path.which_bin(KNOWN_BINARY_NAMES) is None,
        "virtualenv not installed",
    )
    @slowTest
    def test_issue_2068_template_str(self):
        venv_dir = os.path.join(RUNTIME_VARS.TMP, "issue-2068-template-str")

        try:
            ret = self.run_function(
                "state.sls", mods="issue-2068-template-str-no-dot", timeout=120
            )
            self.assertSaltTrueReturn(ret)
        finally:
            if os.path.isdir(venv_dir):
                shutil.rmtree(venv_dir)

        # Let's load the template from the filesystem. If running this state
        # with state.sls works, so should using state.template_str
        template_path = os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            "files",
            "file",
            "base",
            "issue-2068-template-str-no-dot.sls",
        )

        with salt.utils.files.fopen(template_path, "r") as fp_:
            template = salt.utils.stringutils.to_unicode(fp_.read())
            ret = self.run_function("state.template_str", [template], timeout=120)
            self.assertSaltTrueReturn(ret)

        # Now using state.template
        ret = self.run_function("state.template", [template_path], timeout=120)
        self.assertSaltTrueReturn(ret)

        # Now the problematic #2068 including dot's
        ret = self.run_function(
            "state.sls", mods="issue-2068-template-str", timeout=120
        )
        self.assertSaltTrueReturn(ret)

        # Let's load the template from the filesystem. If running this state
        # with state.sls works, so should using state.template_str
        template_path = os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            "files",
            "file",
            "base",
            "issue-2068-template-str.sls",
        )

        with salt.utils.files.fopen(template_path, "r") as fp_:
            template = salt.utils.stringutils.to_unicode(fp_.read())
        ret = self.run_function("state.template_str", [template], timeout=120)
        self.assertSaltTrueReturn(ret)

        # Now using state.template
        ret = self.run_function("state.template", [template_path], timeout=120)
        self.assertSaltTrueReturn(ret)

    @slowTest
    def test_template_invalid_items(self):
        TEMPLATE = textwrap.dedent(
            """\
            {0}:
              - issue-2068-template-str

            /tmp/test-template-invalid-items:
              file:
                - managed
                - source: salt://testfile
            """
        )
        for item in ("include", "exclude", "extends"):
            ret = self.run_function("state.template_str", [TEMPLATE.format(item)])
            self.assertTrue(isinstance(ret, list))
            self.assertNotEqual(ret, [])
            self.assertEqual(
                [
                    "The '{0}' declaration found on '<template-str>' is "
                    "invalid when rendering single templates".format(item)
                ],
                ret,
            )

    @slowTest
    def test_pydsl(self):
        """
        Test the basics of the pydsl
        """
        ret = self.run_function("state.sls", mods="pydsl-1")
        self.assertSaltTrueReturn(ret)

    @slowTest
    def test_issues_7905_and_8174_sls_syntax_error(self):
        """
        Call sls file with yaml syntax error.

        Ensure theses errors are detected and presented to the user without
        stack traces.
        """
        ret = self.run_function("state.sls", mods="syntax.badlist")
        self.assertEqual(
            ret, ["State 'A' in SLS 'syntax.badlist' is not formed as a list"]
        )
        ret = self.run_function("state.sls", mods="syntax.badlist2")
        self.assertEqual(
            ret, ["State 'C' in SLS 'syntax.badlist2' is not formed as a list"]
        )

    @slowTest
    def test_requisites_mixed_require_prereq_use(self):
        """
        Call sls file containing several requisites.
        """
        expected_simple_result = {
            "cmd_|-A_|-echo A_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo A" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo B" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo C" run',
                "result": True,
                "changes": True,
            },
        }
        expected_result = {
            "cmd_|-A_|-echo A fifth_|-run": {
                "__run_num__": 4,
                "comment": 'Command "echo A fifth" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B third_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo B third" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C second_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo C second" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-D_|-echo D first_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo D first" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-E_|-echo E fourth_|-run": {
                "__run_num__": 3,
                "comment": 'Command "echo E fourth" run',
                "result": True,
                "changes": True,
            },
        }
        expected_req_use_result = {
            "cmd_|-A_|-echo A_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo A" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B_|-run": {
                "__run_num__": 4,
                "comment": 'Command "echo B" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo C" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-D_|-echo D_|-run": {
                "__run_num__": 5,
                "comment": 'Command "echo D" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-E_|-echo E_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo E" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-F_|-echo F_|-run": {
                "__run_num__": 3,
                "comment": 'Command "echo F" run',
                "result": True,
                "changes": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.mixed_simple")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_simple_result, result)

        # test Traceback recursion prereq+require #8785
        # TODO: this is actually failing badly
        # ret = self.run_function('state.sls', mods='requisites.prereq_require_recursion_error2')
        # self.assertEqual(
        #    ret,
        #    ['A recursive requisite was found, SLS "requisites.prereq_require_recursion_error2" ID "B" ID "A"']
        # )

        # test Infinite recursion prereq+require #8785 v2
        # TODO: this is actually failing badly
        # ret = self.run_function('state.sls', mods='requisites.prereq_require_recursion_error3')
        # self.assertEqual(
        #    ret,
        #    ['A recursive requisite was found, SLS "requisites.prereq_require_recursion_error2" ID "B" ID "A"']
        # )

        # test Infinite recursion prereq+require #8785 v3
        # TODO: this is actually failing badly, and expected result is maybe not a recursion
        # ret = self.run_function('state.sls', mods='requisites.prereq_require_recursion_error4')
        # self.assertEqual(
        #    ret,
        #    ['A recursive requisite was found, SLS "requisites.prereq_require_recursion_error2" ID "B" ID "A"']
        # )

        # undetected infinite loopS prevents this test from running...
        # TODO: this is actually failing badly
        # ret = self.run_function('state.sls', mods='requisites.mixed_complex1')
        # result = self.normalize_ret(ret)
        # self.assertEqual(expected_result, result)

    @slowTest
    def test_watch_in(self):
        """
        test watch_in requisite when there is a success
        """
        ret = self.run_function("state.sls", mods="requisites.watch_in")
        changes = "test_|-return_changes_|-return_changes_|-succeed_with_changes"
        watch = "test_|-watch_states_|-watch_states_|-succeed_without_changes"

        self.assertEqual(ret[changes]["__run_num__"], 0)
        self.assertEqual(ret[watch]["__run_num__"], 2)

        self.assertEqual("Watch statement fired.", ret[watch]["comment"])
        self.assertEqual(
            "Something pretended to change", ret[changes]["changes"]["testing"]["new"]
        )

    @slowTest
    def test_watch_in_failure(self):
        """
        test watch_in requisite when there is a failure
        """
        ret = self.run_function("state.sls", mods="requisites.watch_in_failure")
        fail = "test_|-return_changes_|-return_changes_|-fail_with_changes"
        watch = "test_|-watch_states_|-watch_states_|-succeed_without_changes"

        self.assertEqual(False, ret[fail]["result"])
        self.assertEqual(
            "One or more requisite failed: requisites.watch_in_failure.return_changes",
            ret[watch]["comment"],
        )

    def normalize_ret(self, ret):
        """
        Normalize the return to the format that we'll use for result checking
        """
        result = {}
        for item, descr in six.iteritems(ret):
            result[item] = {
                "__run_num__": descr["__run_num__"],
                "comment": descr["comment"],
                "result": descr["result"],
                "changes": descr["changes"] != {},  # whether there where any changes
            }
        return result

    @slowTest
    def test_requisites_require_ordering_and_errors(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        expected_result = {
            "cmd_|-A_|-echo A fifth_|-run": {
                "__run_num__": 4,
                "comment": 'Command "echo A fifth" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B second_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo B second" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C third_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo C third" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-D_|-echo D first_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo D first" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-E_|-echo E fourth_|-run": {
                "__run_num__": 3,
                "comment": 'Command "echo E fourth" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-F_|-echo F_|-run": {
                "__run_num__": 5,
                "comment": "The following requisites were not found:\n"
                + "                   require:\n"
                + "                       foobar: A\n",
                "result": False,
                "changes": False,
            },
            "cmd_|-G_|-echo G_|-run": {
                "__run_num__": 6,
                "comment": "The following requisites were not found:\n"
                + "                   require:\n"
                + "                       cmd: Z\n",
                "result": False,
                "changes": False,
            },
            "cmd_|-H_|-echo H_|-run": {
                "__run_num__": 7,
                "comment": "The following requisites were not found:\n"
                + "                   require:\n"
                + "                       cmd: Z\n",
                "result": False,
                "changes": False,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.require")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result, result)

        ret = self.run_function("state.sls", mods="requisites.require_error1")
        self.assertEqual(
            ret,
            [
                "Cannot extend ID 'W' in 'base:requisites.require_error1'. It is not part of the high state.\nThis is likely due to a missing include statement or an incorrectly typed ID.\nEnsure that a state with an ID of 'W' is available\nin environment 'base' and to SLS 'requisites.require_error1'"
            ],
        )

        # issue #8235
        # FIXME: Why is require enforcing list syntax while require_in does not?
        # And why preventing it?
        # Currently this state fails, should return C/B/A
        result = {}
        ret = self.run_function("state.sls", mods="requisites.require_simple_nolist")
        self.assertEqual(
            ret,
            [
                "The require statement in state 'B' in SLS "
                + "'requisites.require_simple_nolist' needs to be formed as a list"
            ],
        )

        # commented until a fix is made for issue #8772
        # TODO: this test actually fails
        # ret = self.run_function('state.sls', mods='requisites.require_error2')
        # self.assertEqual(ret, [
        #    'Cannot extend state foobar for ID A in "base:requisites.require_error2".'
        #    + ' It is not part of the high state.'
        # ])

        ret = self.run_function("state.sls", mods="requisites.require_recursion_error1")
        self.assertEqual(
            ret,
            [
                'A recursive requisite was found, SLS "requisites.require_recursion_error1" ID "B" ID "A"'
            ],
        )

    @slowTest
    def test_requisites_require_any(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        expected_result = {
            "cmd_|-A_|-echo A_|-run": {
                "__run_num__": 3,
                "comment": 'Command "echo A" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo B" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-$(which false)_|-run": {
                "__run_num__": 1,
                "comment": 'Command "$(which false)" run',
                "result": False,
                "changes": True,
            },
            "cmd_|-D_|-echo D_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo D" run',
                "result": True,
                "changes": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.require_any")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result, result)

    @slowTest
    def test_requisites_require_any_fail(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        ret = self.run_function("state.sls", mods="requisites.require_any_fail")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertIn(
            "One or more requisite failed", result["cmd_|-D_|-echo D_|-run"]["comment"]
        )

    @slowTest
    def test_requisites_watch_any(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        if salt.utils.platform.is_windows():
            cmd_true = "exit"
            cmd_false = "exit /B 1"
        else:
            cmd_true = "true"
            cmd_false = "false"
        expected_result = {
            "cmd_|-A_|-{0}_|-wait".format(cmd_true): {
                "__run_num__": 4,
                "comment": 'Command "{0}" run'.format(cmd_true),
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-{0}_|-run".format(cmd_true): {
                "__run_num__": 0,
                "comment": 'Command "{0}" run'.format(cmd_true),
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-{0}_|-run".format(cmd_false): {
                "__run_num__": 1,
                "comment": 'Command "{0}" run'.format(cmd_false),
                "result": False,
                "changes": True,
            },
            "cmd_|-D_|-{0}_|-run".format(cmd_true): {
                "__run_num__": 2,
                "comment": 'Command "{0}" run'.format(cmd_true),
                "result": True,
                "changes": True,
            },
            "cmd_|-E_|-{0}_|-wait".format(cmd_true): {
                "__run_num__": 9,
                "comment": 'Command "{0}" run'.format(cmd_true),
                "result": True,
                "changes": True,
            },
            "cmd_|-F_|-{0}_|-run".format(cmd_true): {
                "__run_num__": 5,
                "comment": 'Command "{0}" run'.format(cmd_true),
                "result": True,
                "changes": True,
            },
            "cmd_|-G_|-{0}_|-run".format(cmd_false): {
                "__run_num__": 6,
                "comment": 'Command "{0}" run'.format(cmd_false),
                "result": False,
                "changes": True,
            },
            "cmd_|-H_|-{0}_|-run".format(cmd_false): {
                "__run_num__": 7,
                "comment": 'Command "{0}" run'.format(cmd_false),
                "result": False,
                "changes": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.watch_any")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result, result)

    @slowTest
    def test_requisites_watch_any_fail(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        ret = self.run_function("state.sls", mods="requisites.watch_any_fail")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertIn(
            "One or more requisite failed", result["cmd_|-A_|-true_|-wait"]["comment"]
        )

    @slowTest
    def test_requisites_onchanges_any(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        expected_result = {
            'cmd_|-another_changing_state_|-echo "Changed!"_|-run': {
                "__run_num__": 1,
                "changes": True,
                "comment": 'Command "echo "Changed!"" run',
                "result": True,
            },
            'cmd_|-changing_state_|-echo "Changed!"_|-run': {
                "__run_num__": 0,
                "changes": True,
                "comment": 'Command "echo "Changed!"" run',
                "result": True,
            },
            'cmd_|-test_one_changing_states_|-echo "Success!"_|-run': {
                "__run_num__": 4,
                "changes": True,
                "comment": 'Command "echo "Success!"" run',
                "result": True,
            },
            'cmd_|-test_two_non_changing_states_|-echo "Should not run"_|-run': {
                "__run_num__": 5,
                "changes": False,
                "comment": "State was not run because none of the onchanges reqs changed",
                "result": True,
            },
            "pip_|-another_non_changing_state_|-mock_|-installed": {
                "__run_num__": 3,
                "changes": False,
                "comment": "Python package mock was already installed\nAll specified packages are already installed",
                "result": True,
            },
            "pip_|-non_changing_state_|-mock_|-installed": {
                "__run_num__": 2,
                "changes": False,
                "comment": "Python package mock was already installed\nAll specified packages are already installed",
                "result": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.onchanges_any")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result, result)

    @slowTest
    def test_requisites_onfail_any(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        expected_result = {
            "cmd_|-a_|-exit 0_|-run": {
                "__run_num__": 0,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-b_|-exit 1_|-run": {
                "__run_num__": 1,
                "changes": True,
                "comment": 'Command "exit 1" run',
                "result": False,
            },
            "cmd_|-c_|-exit 0_|-run": {
                "__run_num__": 2,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-d_|-echo itworked_|-run": {
                "__run_num__": 3,
                "changes": True,
                "comment": 'Command "echo itworked" run',
                "result": True,
            },
            "cmd_|-e_|-exit 0_|-run": {
                "__run_num__": 4,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-f_|-exit 0_|-run": {
                "__run_num__": 5,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-g_|-exit 0_|-run": {
                "__run_num__": 6,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-h_|-echo itworked_|-run": {
                "__run_num__": 7,
                "changes": False,
                "comment": "State was not run because onfail req did not change",
                "result": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.onfail_any")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result, result)

    @slowTest
    def test_requisites_onfail_all(self):
        """
        Call sls file containing several onfail-all

        Ensure that some of them are failing and that the order is right.
        """
        expected_result = {
            "cmd_|-a_|-exit 0_|-run": {
                "__run_num__": 0,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-b_|-exit 0_|-run": {
                "__run_num__": 1,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-c_|-exit 0_|-run": {
                "__run_num__": 2,
                "changes": True,
                "comment": 'Command "exit 0" run',
                "result": True,
            },
            "cmd_|-d_|-exit 1_|-run": {
                "__run_num__": 3,
                "changes": True,
                "comment": 'Command "exit 1" run',
                "result": False,
            },
            "cmd_|-e_|-exit 1_|-run": {
                "__run_num__": 4,
                "changes": True,
                "comment": 'Command "exit 1" run',
                "result": False,
            },
            "cmd_|-f_|-exit 1_|-run": {
                "__run_num__": 5,
                "changes": True,
                "comment": 'Command "exit 1" run',
                "result": False,
            },
            "cmd_|-reqs also met_|-echo itonfailed_|-run": {
                "__run_num__": 9,
                "changes": True,
                "comment": 'Command "echo itonfailed" run',
                "result": True,
            },
            "cmd_|-reqs also not met_|-echo italsodidnonfail_|-run": {
                "__run_num__": 7,
                "changes": False,
                "comment": "State was not run because onfail req did not change",
                "result": True,
            },
            "cmd_|-reqs met_|-echo itonfailed_|-run": {
                "__run_num__": 8,
                "changes": True,
                "comment": 'Command "echo itonfailed" run',
                "result": True,
            },
            "cmd_|-reqs not met_|-echo itdidntonfail_|-run": {
                "__run_num__": 6,
                "changes": False,
                "comment": "State was not run because onfail req did not change",
                "result": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.onfail_all")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result, result)

    @slowTest
    def test_requisites_full_sls(self):
        """
        Teste the sls special command in requisites
        """
        expected_result = {
            "cmd_|-A_|-echo A_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo A" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo B" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo C" run',
                "result": True,
                "changes": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.fullsls_require")
        self.assertReturnNonEmptySaltType(ret)
        result = self.normalize_ret(ret)
        self.assertEqual(expected_result, result)

        # issue #8233: traceback on prereq sls
        # TODO: not done
        # ret = self.run_function('state.sls', mods='requisites.fullsls_prereq')
        # self.assertEqual(['sls command can only be used with require requisite'], ret)

    @slowTest
    def test_requisites_require_no_state_module(self):
        """
        Call sls file containing several require_in and require.

        Ensure that some of them are failing and that the order is right.
        """
        expected_result = {
            "cmd_|-A_|-echo A fifth_|-run": {
                "__run_num__": 4,
                "comment": 'Command "echo A fifth" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B second_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo B second" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C third_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo C third" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-D_|-echo D first_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo D first" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-E_|-echo E fourth_|-run": {
                "__run_num__": 3,
                "comment": 'Command "echo E fourth" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-G_|-echo G_|-run": {
                "__run_num__": 5,
                "comment": "The following requisites were not found:\n"
                + "                   require:\n"
                + "                       id: Z\n",
                "result": False,
                "changes": False,
            },
            "cmd_|-H_|-echo H_|-run": {
                "__run_num__": 6,
                "comment": "The following requisites were not found:\n"
                + "                   require:\n"
                + "                       id: Z\n",
                "result": False,
                "changes": False,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.require_no_state_module")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result, result)

    @slowTest
    def test_requisites_prereq_simple_ordering_and_errors(self):
        """
        Call sls file containing several prereq_in and prereq.

        Ensure that some of them are failing and that the order is right.
        """
        expected_result_simple = {
            "cmd_|-A_|-echo A third_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo A third" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B first_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo B first" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C second_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo C second" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-I_|-echo I_|-run": {
                "__run_num__": 3,
                "comment": "The following requisites were not found:\n"
                + "                   prereq:\n"
                + "                       cmd: Z\n",
                "result": False,
                "changes": False,
            },
            "cmd_|-J_|-echo J_|-run": {
                "__run_num__": 4,
                "comment": "The following requisites were not found:\n"
                + "                   prereq:\n"
                + "                       foobar: A\n",
                "result": False,
                "changes": False,
            },
        }
        expected_result_simple_no_state_module = {
            "cmd_|-A_|-echo A third_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo A third" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B first_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo B first" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C second_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo C second" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-I_|-echo I_|-run": {
                "__run_num__": 3,
                "comment": "The following requisites were not found:\n"
                + "                   prereq:\n"
                + "                       id: Z\n",
                "result": False,
                "changes": False,
            },
        }
        expected_result_simple2 = {
            "cmd_|-A_|-echo A_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo A" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo B" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo C" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-D_|-echo D_|-run": {
                "__run_num__": 3,
                "comment": 'Command "echo D" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-E_|-echo E_|-run": {
                "__run_num__": 4,
                "comment": 'Command "echo E" run',
                "result": True,
                "changes": True,
            },
        }
        expected_result_simple3 = {
            "cmd_|-A_|-echo A first_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo A first" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B second_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo B second" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C third_|-wait": {
                "__run_num__": 2,
                "comment": "",
                "result": True,
                "changes": False,
            },
        }
        expected_result_complex = {
            "cmd_|-A_|-echo A fourth_|-run": {
                "__run_num__": 3,
                "comment": 'Command "echo A fourth" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-B_|-echo B first_|-run": {
                "__run_num__": 0,
                "comment": 'Command "echo B first" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-C_|-echo C second_|-run": {
                "__run_num__": 1,
                "comment": 'Command "echo C second" run',
                "result": True,
                "changes": True,
            },
            "cmd_|-D_|-echo D third_|-run": {
                "__run_num__": 2,
                "comment": 'Command "echo D third" run',
                "result": True,
                "changes": True,
            },
        }
        ret = self.run_function("state.sls", mods="requisites.prereq_simple")
        self.assertReturnNonEmptySaltType(ret)
        result = self.normalize_ret(ret)
        self.assertEqual(expected_result_simple, result)

        # same test, but not using lists in yaml syntax
        # TODO: issue #8235, prereq ignored when not used in list syntax
        # Currently fails badly with :
        # TypeError encountered executing state.sls: string indices must be integers, not str.
        # expected_result_simple.pop('cmd_|-I_|-echo I_|-run')
        # expected_result_simple.pop('cmd_|-J_|-echo J_|-run')
        # ret = self.run_function('state.sls', mods='requisites.prereq_simple_nolist')
        # result = self.normalize_ret(ret)
        # self.assertEqual(expected_result_simple, result)

        ret = self.run_function("state.sls", mods="requisites.prereq_simple2")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result_simple2, result)

        ret = self.run_function("state.sls", mods="requisites.prereq_simple3")
        result = self.normalize_ret(ret)
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(expected_result_simple3, result)

        # ret = self.run_function('state.sls', mods='requisites.prereq_error_nolist')
        # self.assertEqual(
        #    ret,
        #    ['Cannot extend ID Z in "base:requisites.prereq_error_nolist".'
        #    + ' It is not part of the high state.']
        # )

        ret = self.run_function("state.sls", mods="requisites.prereq_compile_error1")
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(
            ret["cmd_|-B_|-echo B_|-run"]["comment"],
            "The following requisites were not found:\n"
            + "                   prereq:\n"
            + "                       foobar: A\n",
        )

        ret = self.run_function("state.sls", mods="requisites.prereq_compile_error2")
        self.assertReturnNonEmptySaltType(ret)
        self.assertEqual(
            ret["cmd_|-B_|-echo B_|-run"]["comment"],
            "The following requisites were not found:\n"
            + "                   prereq:\n"
            + "                       foobar: C\n",
        )

        ret = self.run_function("state.sls", mods="requisites.prereq_complex")
        result = self.normalize_ret(ret)
        self.assertEqual(expected_result_complex, result)

        # issue #8210 : prereq recursion undetected
        # TODO: this test fails
        # ret = self.run_function('state.sls', mods='requisites.prereq_recursion_error')
        # self.assertEqual(
        #    ret,
        #    ['A recursive requisite was found, SLS "requisites.prereq_recursion_error" ID "B" ID "A"']
        # )

        ret = self.run_function(
            "state.sls", mods="requisites.prereq_simple_no_state_module"
        )
        result = self.normalize_ret(ret)
        self.assertEqual(expected_result_simple_no_state_module, result)

    @slowTest
    def test_infinite_recursion_sls_prereq(self):
        ret = self.run_function(
            "state.sls", mods="requisites.prereq_sls_infinite_recursion"
        )
        self.assertSaltTrueReturn(ret)

    @slowTest
    def test_requisites_use(self):
        """
        Call sls file containing several use_in and use.

        """
        # TODO issue #8235 & #8774 some examples are still commented in the test file
        ret = self.run_function("state.sls", mods="requisites.use")
        self.assertReturnNonEmptySaltType(ret)
        for item, descr in six.iteritems(ret):
            self.assertEqual(descr["comment"], "onlyif condition is false")

        # TODO: issue #8802 : use recursions undetected
        # issue is closed as use does not actually inherit requisites
        # if chain-use is added after #8774 resolution theses tests would maybe become useful
        # ret = self.run_function('state.sls', mods='requisites.use_recursion')
        # self.assertEqual(ret, [
        #    'A recursive requisite was found, SLS "requisites.use_recursion"'
        #    + ' ID "B" ID "A"'
        # ])

        # ret = self.run_function('state.sls', mods='requisites.use_recursion2')
        # self.assertEqual(ret, [
        #    'A recursive requisite was found, SLS "requisites.use_recursion2"'
        #    + ' ID "C" ID "A"'
        # ])

        # ret = self.run_function('state.sls', mods='requisites.use_auto_recursion')
        # self.assertEqual(ret, [
        #    'A recursive requisite was found, SLS "requisites.use_recursion"'
        #    + ' ID "A" ID "A"'
        # ])

    @slowTest
    def test_requisites_use_no_state_module(self):
        """
        Call sls file containing several use_in and use.

        """
        ret = self.run_function("state.sls", mods="requisites.use_no_state_module")
        self.assertReturnNonEmptySaltType(ret)
        for item, descr in six.iteritems(ret):
            self.assertEqual(descr["comment"], "onlyif condition is false")

    @slowTest
    def test_onlyif_req(self):
        ret = self.run_function(
            "state.single",
            fun="test.succeed_with_changes",
            name="onlyif test",
            onlyif=[{}],
        )["test_|-onlyif test_|-onlyif test_|-succeed_with_changes"]
        self.assertTrue(ret["result"])
        self.assertEqual(ret["comment"], "Success!")
        ret = self.run_function(
            "state.single",
            fun="test.fail_with_changes",
            name="onlyif test",
            onlyif=[{"fun": "test.false"}],
        )["test_|-onlyif test_|-onlyif test_|-fail_with_changes"]
        self.assertTrue(ret["result"])
        self.assertFalse(ret["changes"])
        self.assertEqual(ret["comment"], "onlyif condition is false")
        ret = self.run_function(
            "state.single",
            fun="test.fail_with_changes",
            name="onlyif test",
            onlyif=[{"fun": "test.true"}],
        )["test_|-onlyif test_|-onlyif test_|-fail_with_changes"]
        self.assertFalse(ret["result"])
        self.assertTrue(ret["changes"])
        self.assertEqual(ret["comment"], "Failure!")
        ret = self.run_function(
            "state.single",
            fun="test.succeed_without_changes",
            name="onlyif test",
            onlyif=[{"fun": "test.true"}],
        )["test_|-onlyif test_|-onlyif test_|-succeed_without_changes"]
        self.assertTrue(ret["result"])
        self.assertFalse(ret["changes"])
        self.assertEqual(ret["comment"], "Success!")

    @slowTest
    def test_onlyif_req_retcode(self):
        ret = self.run_function(
            "state.single",
            fun="test.succeed_with_changes",
            name="onlyif test",
            onlyif=[{"fun": "test.retcode"}],
        )["test_|-onlyif test_|-onlyif test_|-succeed_with_changes"]
        self.assertTrue(ret["result"])
        self.assertFalse(ret["changes"])
        self.assertEqual(ret["comment"], "onlyif condition is false")
        ret = self.run_function(
            "state.single",
            fun="test.succeed_with_changes",
            name="onlyif test",
            onlyif=[{"fun": "test.retcode", "code": 0}],
        )["test_|-onlyif test_|-onlyif test_|-succeed_with_changes"]
        self.assertTrue(ret["result"])
        self.assertTrue(ret["changes"])
        self.assertEqual(ret["comment"], "Success!")

    @slowTest
    def test_unless_req(self):
        ret = self.run_function(
            "state.single",
            fun="test.succeed_with_changes",
            name="unless test",
            unless=[{}],
        )["test_|-unless test_|-unless test_|-succeed_with_changes"]
        self.assertTrue(ret["result"])
        self.assertEqual(ret["comment"], "Success!")
        ret = self.run_function(
            "state.single",
            fun="test.fail_with_changes",
            name="unless test",
            unless=[{"fun": "test.true"}],
        )["test_|-unless test_|-unless test_|-fail_with_changes"]
        self.assertTrue(ret["result"])
        self.assertFalse(ret["changes"])
        self.assertEqual(ret["comment"], "unless condition is true")
        ret = self.run_function(
            "state.single",
            fun="test.fail_with_changes",
            name="unless test",
            unless=[{"fun": "test.false"}],
        )["test_|-unless test_|-unless test_|-fail_with_changes"]
        self.assertFalse(ret["result"])
        self.assertTrue(ret["changes"])
        self.assertEqual(ret["comment"], "Failure!")
        ret = self.run_function(
            "state.single",
            fun="test.succeed_without_changes",
            name="unless test",
            unless=[{"fun": "test.false"}],
        )["test_|-unless test_|-unless test_|-succeed_without_changes"]
        self.assertTrue(ret["result"])
        self.assertFalse(ret["changes"])
        self.assertEqual(ret["comment"], "Success!")

    @slowTest
    def test_unless_req_retcode(self):
        ret = self.run_function(
            "state.single",
            fun="test.succeed_with_changes",
            name="unless test",
            unless=[{"fun": "test.retcode"}],
        )["test_|-unless test_|-unless test_|-succeed_with_changes"]
        self.assertTrue(ret["result"])
        self.assertTrue(ret["changes"])
        self.assertEqual(ret["comment"], "Success!")
        ret = self.run_function(
            "state.single",
            fun="test.succeed_with_changes",
            name="unless test",
            unless=[{"fun": "test.retcode", "code": 0}],
        )["test_|-unless test_|-unless test_|-succeed_with_changes"]
        self.assertTrue(ret["result"])
        self.assertFalse(ret["changes"])
        self.assertEqual(ret["comment"], "unless condition is true")

    @slowTest
    @pytest.mark.usefixtures("salt_sub_minion")
    def test_get_file_from_env_in_top_match(self):
        tgt = os.path.join(RUNTIME_VARS.TMP, "prod-cheese-file")
        try:
            ret = self.run_function("state.highstate", minion_tgt="sub_minion")
            self.assertSaltTrueReturn(ret)
            self.assertTrue(os.path.isfile(tgt))
            with salt.utils.files.fopen(tgt, "r") as cheese:
                data = salt.utils.stringutils.to_unicode(cheese.read())
            self.assertIn("Gromit", data)
            self.assertIn("Comte", data)
        finally:
            if os.path.islink(tgt):
                os.unlink(tgt)

    # onchanges tests

    @slowTest
    def test_onchanges_requisite(self):
        """
        Tests a simple state using the onchanges requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.onchanges_simple")

        # First, test the result of the state run when changes are expected to happen
        test_data = state_run['cmd_|-test_changing_state_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

        # Then, test the result of the state run when changes are not expected to happen
        test_data = state_run[
            'cmd_|-test_non_changing_state_|-echo "Should not run"_|-run'
        ]["comment"]
        expected_result = "State was not run because none of the onchanges reqs changed"
        self.assertIn(expected_result, test_data)

    @slowTest
    def test_onchanges_requisite_multiple(self):
        """
        Tests a simple state using the onchanges requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.onchanges_multiple")

        # First, test the result of the state run when two changes are expected to happen
        test_data = state_run['cmd_|-test_two_changing_states_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

        # Then, test the result of the state run when two changes are not expected to happen
        test_data = state_run[
            'cmd_|-test_two_non_changing_states_|-echo "Should not run"_|-run'
        ]["comment"]
        expected_result = "State was not run because none of the onchanges reqs changed"
        self.assertIn(expected_result, test_data)

        # Finally, test the result of the state run when only one of the onchanges requisites changes.
        test_data = state_run['cmd_|-test_one_changing_state_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

    @slowTest
    def test_onchanges_in_requisite(self):
        """
        Tests a simple state using the onchanges_in requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function(
            "state.sls", mods="requisites.onchanges_in_simple"
        )

        # First, test the result of the state run of when changes are expected to happen
        test_data = state_run['cmd_|-test_changes_expected_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

        # Then, test the result of the state run when changes are not expected to happen
        test_data = state_run[
            'cmd_|-test_changes_not_expected_|-echo "Should not run"_|-run'
        ]["comment"]
        expected_result = "State was not run because none of the onchanges reqs changed"
        self.assertIn(expected_result, test_data)

    @slowTest
    def test_onchanges_requisite_no_state_module(self):
        """
        Tests a simple state using the onchanges requisite without state modules
        """
        # Only run the state once and keep the return data
        state_run = self.run_function(
            "state.sls", mods="requisites.onchanges_simple_no_state_module"
        )
        test_data = state_run['cmd_|-test_changing_state_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

    @slowTest
    def test_onchanges_requisite_with_duration(self):
        """
        Tests a simple state using the onchanges requisite
        the state will not run but results will include duration
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.onchanges_simple")

        # Then, test the result of the state run when changes are not expected to happen
        # and ensure duration is included in the results
        test_data = state_run[
            'cmd_|-test_non_changing_state_|-echo "Should not run"_|-run'
        ]
        self.assertIn("duration", test_data)

    # onfail tests

    @slowTest
    def test_onfail_requisite(self):
        """
        Tests a simple state using the onfail requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.onfail_simple")

        # First, test the result of the state run when a failure is expected to happen
        test_data = state_run['cmd_|-test_failing_state_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

        # Then, test the result of the state run when a failure is not expected to happen
        test_data = state_run[
            'cmd_|-test_non_failing_state_|-echo "Should not run"_|-run'
        ]["comment"]
        expected_result = "State was not run because onfail req did not change"
        self.assertIn(expected_result, test_data)

    @slowTest
    def test_multiple_onfail_requisite(self):
        """
        test to ensure state is run even if only one
        of the onfails fails. This is a test for the issue:
        https://github.com/saltstack/salt/issues/22370
        """

        state_run = self.run_function(
            "state.sls", mods="requisites.onfail_multiple", timeout=self.TIMEOUT
        )

        retcode = state_run["cmd_|-c_|-echo itworked_|-run"]["changes"]["retcode"]
        self.assertEqual(retcode, 0)

        stdout = state_run["cmd_|-c_|-echo itworked_|-run"]["changes"]["stdout"]
        self.assertEqual(stdout, "itworked")

    @slowTest
    def test_onfail_in_requisite(self):
        """
        Tests a simple state using the onfail_in requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.onfail_in_simple")

        # First, test the result of the state run when a failure is expected to happen
        test_data = state_run['cmd_|-test_failing_state_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

        # Then, test the result of the state run when a failure is not expected to happen
        test_data = state_run[
            'cmd_|-test_non_failing_state_|-echo "Should not run"_|-run'
        ]["comment"]
        expected_result = "State was not run because onfail req did not change"
        self.assertIn(expected_result, test_data)

    @slowTest
    def test_onfail_requisite_no_state_module(self):
        """
        Tests a simple state using the onfail requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function(
            "state.sls", mods="requisites.onfail_simple_no_state_module"
        )

        # First, test the result of the state run when a failure is expected to happen
        test_data = state_run['cmd_|-test_failing_state_|-echo "Success!"_|-run'][
            "comment"
        ]
        expected_result = 'Command "echo "Success!"" run'
        self.assertIn(expected_result, test_data)

        # Then, test the result of the state run when a failure is not expected to happen
        test_data = state_run[
            'cmd_|-test_non_failing_state_|-echo "Should not run"_|-run'
        ]["comment"]
        expected_result = "State was not run because onfail req did not change"
        self.assertIn(expected_result, test_data)

    @slowTest
    def test_onfail_requisite_with_duration(self):
        """
        Tests a simple state using the onfail requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.onfail_simple")

        # Then, test the result of the state run when a failure is not expected to happen
        test_data = state_run[
            'cmd_|-test_non_failing_state_|-echo "Should not run"_|-run'
        ]
        self.assertIn("duration", test_data)

    @slowTest
    def test_multiple_onfail_requisite_with_required(self):
        """
        test to ensure multiple states are run
        when specified as onfails for a single state.
        This is a test for the issue:
        https://github.com/saltstack/salt/issues/46552
        """

        state_run = self.run_function(
            "state.sls", mods="requisites.onfail_multiple_required"
        )

        retcode = state_run["cmd_|-b_|-echo b_|-run"]["changes"]["retcode"]
        self.assertEqual(retcode, 0)

        retcode = state_run["cmd_|-c_|-echo c_|-run"]["changes"]["retcode"]
        self.assertEqual(retcode, 0)

        retcode = state_run["cmd_|-d_|-echo d_|-run"]["changes"]["retcode"]
        self.assertEqual(retcode, 0)

        stdout = state_run["cmd_|-b_|-echo b_|-run"]["changes"]["stdout"]
        self.assertEqual(stdout, "b")

        stdout = state_run["cmd_|-c_|-echo c_|-run"]["changes"]["stdout"]
        self.assertEqual(stdout, "c")

        stdout = state_run["cmd_|-d_|-echo d_|-run"]["changes"]["stdout"]
        self.assertEqual(stdout, "d")

        comment = state_run["cmd_|-e_|-echo e_|-run"]["comment"]
        self.assertEqual(comment, "State was not run because onfail req did not change")

        stdout = state_run["cmd_|-f_|-echo f_|-run"]["changes"]["stdout"]
        self.assertEqual(stdout, "f")

    @slowTest
    def test_multiple_onfail_requisite_with_required_no_run(self):
        """
        test to ensure multiple states are not run
        when specified as onfails for a single state
        which fails.
        This is a test for the issue:
        https://github.com/saltstack/salt/issues/46552
        """

        state_run = self.run_function(
            "state.sls", mods="requisites.onfail_multiple_required_no_run"
        )

        expected = "State was not run because onfail req did not change"

        stdout = state_run["cmd_|-b_|-echo b_|-run"]["comment"]
        self.assertEqual(stdout, expected)

        stdout = state_run["cmd_|-c_|-echo c_|-run"]["comment"]
        self.assertEqual(stdout, expected)

        stdout = state_run["cmd_|-d_|-echo d_|-run"]["comment"]
        self.assertEqual(stdout, expected)

    # listen tests

    @slowTest
    def test_listen_requisite(self):
        """
        Tests a simple state using the listen requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.listen_simple")

        # First, test the result of the state run when a listener is expected to trigger
        listener_state = 'cmd_|-listener_test_listening_change_state_|-echo "Listening State"_|-mod_watch'
        self.assertIn(listener_state, state_run)

        # Then, test the result of the state run when a listener should not trigger
        absent_state = 'cmd_|-listener_test_listening_non_changing_state_|-echo "Only run once"_|-mod_watch'
        self.assertNotIn(absent_state, state_run)

    @slowTest
    def test_listen_in_requisite(self):
        """
        Tests a simple state using the listen_in requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.listen_in_simple")

        # First, test the result of the state run when a listener is expected to trigger
        listener_state = 'cmd_|-listener_test_listening_change_state_|-echo "Listening State"_|-mod_watch'
        self.assertIn(listener_state, state_run)

        # Then, test the result of the state run when a listener should not trigger
        absent_state = 'cmd_|-listener_test_listening_non_changing_state_|-echo "Only run once"_|-mod_watch'
        self.assertNotIn(absent_state, state_run)

    @slowTest
    def test_listen_in_requisite_resolution(self):
        """
        Verify listen_in requisite lookups use ID declaration to check for changes
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.listen_in_simple")

        # Test the result of the state run when a listener is expected to trigger
        listener_state = 'cmd_|-listener_test_listen_in_resolution_|-echo "Successful listen_in resolution"_|-mod_watch'
        self.assertIn(listener_state, state_run)

    @slowTest
    def test_listen_requisite_resolution(self):
        """
        Verify listen requisite lookups use ID declaration to check for changes
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.listen_simple")

        # Both listeners are expected to trigger
        listener_state = 'cmd_|-listener_test_listening_resolution_one_|-echo "Successful listen resolution"_|-mod_watch'
        self.assertIn(listener_state, state_run)

        listener_state = 'cmd_|-listener_test_listening_resolution_two_|-echo "Successful listen resolution"_|-mod_watch'
        self.assertIn(listener_state, state_run)

    @slowTest
    def test_listen_requisite_no_state_module(self):
        """
        Tests a simple state using the listen requisite
        """

        # Only run the state once and keep the return data
        state_run = self.run_function(
            "state.sls", mods="requisites.listen_simple_no_state_module"
        )
        # First, test the result of the state run when a listener is expected to trigger
        listener_state = 'cmd_|-listener_test_listening_change_state_|-echo "Listening State"_|-mod_watch'
        self.assertIn(listener_state, state_run)

        # Then, test the result of the state run when a listener should not trigger
        absent_state = 'cmd_|-listener_test_listening_non_changing_state_|-echo "Only run once"_|-mod_watch'
        self.assertNotIn(absent_state, state_run)

    @slowTest
    def test_listen_in_requisite_resolution_names(self):
        """
        Verify listen_in requisite lookups use ID declaration to check for changes
        and resolves magic names state variable
        """

        # Only run the state once and keep the return data
        state_run = self.run_function("state.sls", mods="requisites.listen_in_names")
        self.assertIn("test_|-listener_service_|-nginx_|-mod_watch", state_run)
        self.assertIn("test_|-listener_service_|-crond_|-mod_watch", state_run)

    @slowTest
    def test_listen_requisite_resolution_names(self):
        """
        Verify listen requisite lookups use ID declaration to check for changes
        and resolves magic names state variable
        """

        # Only run the state once and keep the return data
        state_run = self.run_function(
            "state.sls", mods="requisites.listen_names", timeout=self.TIMEOUT
        )
        self.assertIn("test_|-listener_service_|-nginx_|-mod_watch", state_run)
        self.assertIn("test_|-listener_service_|-crond_|-mod_watch", state_run)

    @slowTest
    def test_issue_30820_requisite_in_match_by_name(self):
        """
        This tests the case where a requisite_in matches by name instead of ID

        See https://github.com/saltstack/salt/issues/30820 for more info
        """
        state_run = self.run_function(
            "state.sls", mods="requisites.requisite_in_match_by_name"
        )
        bar_state = "cmd_|-bar state_|-echo bar_|-wait"

        self.assertIn(bar_state, state_run)
        self.assertEqual(state_run[bar_state]["comment"], 'Command "echo bar" run')

    @slowTest
    def test_retry_option_defaults(self):
        """
        test the retry option on a simple state with defaults
        ensure comment is as expected
        ensure state duration is greater than default retry_interval (30 seconds)
        """
        state_run = self.run_function("state.sls", mods="retry.retry_defaults")
        retry_state = "file_|-file_test_|-/path/to/a/non-existent/file.txt_|-exists"
        expected_comment = (
            'Attempt 1: Returned a result of "False", with the following '
            'comment: "Specified path /path/to/a/non-existent/file.txt does not exist"\n'
            "Specified path /path/to/a/non-existent/file.txt does not exist"
        )
        self.assertEqual(state_run[retry_state]["comment"], expected_comment)
        self.assertTrue(state_run[retry_state]["duration"] > 30)
        self.assertEqual(state_run[retry_state]["result"], False)

    @slowTest
    def test_retry_option_custom(self):
        """
        test the retry option on a simple state with custom retry values
        ensure comment is as expected
        ensure state duration is greater than custom defined interval * (retries - 1)
        """
        state_run = self.run_function("state.sls", mods="retry.retry_custom")
        retry_state = "file_|-file_test_|-/path/to/a/non-existent/file.txt_|-exists"
        expected_comment = (
            'Attempt 1: Returned a result of "False", with the following '
            'comment: "Specified path /path/to/a/non-existent/file.txt does not exist"\n'
            'Attempt 2: Returned a result of "False", with the following comment: "Specified'
            ' path /path/to/a/non-existent/file.txt does not exist"\nAttempt 3: Returned'
            ' a result of "False", with the following comment: "Specified path'
            ' /path/to/a/non-existent/file.txt does not exist"\nAttempt 4: Returned a'
            ' result of "False", with the following comment: "Specified path'
            ' /path/to/a/non-existent/file.txt does not exist"\nSpecified path'
            " /path/to/a/non-existent/file.txt does not exist"
        )
        self.assertEqual(state_run[retry_state]["comment"], expected_comment)
        self.assertTrue(state_run[retry_state]["duration"] > 40)
        self.assertEqual(state_run[retry_state]["result"], False)

    @slowTest
    def test_retry_option_success(self):
        """
        test a state with the retry option that should return True immedietly (i.e. no retries)
        """
        testfile = os.path.join(RUNTIME_VARS.TMP, "retry_file_option_success")
        state_run = self.run_function("state.sls", mods="retry.retry_success")
        os.unlink(testfile)
        retry_state = "file_|-file_test_|-{0}_|-exists".format(testfile)
        self.assertNotIn("Attempt", state_run[retry_state]["comment"])

    def run_create(self, testfile):
        """
        helper function to wait 30 seconds and then create the temp retry file
        """
        # Wait for the requisite stae 'file_test_a' to complete before creating
        # test_file
        while True:
            if os.path.exists(testfile + "_a"):
                break
            time.sleep(1)
        time.sleep(30)
        with salt.utils.files.fopen(testfile, "a"):
            pass

    @slowTest
    def test_retry_option_eventual_success(self):
        """
        test a state with the retry option that should return True after at least 4 retry attmempt
        but never run 15 attempts
        """
        testfile = os.path.join(RUNTIME_VARS.TMP, "retry_file_eventual_success")
        assert not os.path.exists(testfile + "_a")
        assert not os.path.exists(testfile)
        create_thread = threading.Thread(target=self.run_create, args=(testfile,))
        create_thread.start()
        state_run = self.run_function("state.sls", mods="retry.retry_success2")
        retry_state = "file_|-file_test_b_|-{0}_|-exists".format(testfile)
        self.assertIn("Attempt 1:", state_run[retry_state]["comment"])
        self.assertIn("Attempt 2:", state_run[retry_state]["comment"])
        self.assertIn("Attempt 3:", state_run[retry_state]["comment"])
        self.assertIn("Attempt 4:", state_run[retry_state]["comment"])
        self.assertNotIn("Attempt 15:", state_run[retry_state]["comment"])
        self.assertEqual(state_run[retry_state]["result"], True)

    @slowTest
    def test_issue_38683_require_order_failhard_combination(self):
        """
        This tests the case where require, order, and failhard are all used together in a state definition.

        Previously, the order option, which used in tandem with require and failhard, would cause the state
        compiler to stacktrace. This exposed a logic error in the ``check_failhard`` function of the state
        compiler. With the logic error resolved, this test should now pass.

        See https://github.com/saltstack/salt/issues/38683 for more information.
        """
        state_run = self.run_function(
            "state.sls", mods="requisites.require_order_failhard_combo"
        )
        state_id = "test_|-b_|-b_|-fail_with_changes"

        self.assertIn(state_id, state_run)
        self.assertEqual(state_run[state_id]["comment"], "Failure!")
        self.assertFalse(state_run[state_id]["result"])

    @slowTest
    def test_issue_46762_prereqs_on_a_state_with_unfulfilled_requirements(self):
        """
        This tests the case where state C requires state A, which fails.
        State C is a pre-required state for State B.
        Since state A fails, state C will not run because the requisite failed,
        therefore state B will not run because state C failed to run.

        See https://github.com/saltstack/salt/issues/46762 for
        more information.
        """
        state_run = self.run_function("state.sls", mods="issue-46762")

        state_id = "test_|-a_|-a_|-fail_without_changes"
        self.assertIn(state_id, state_run)
        self.assertEqual(state_run[state_id]["comment"], "Failure!")
        self.assertFalse(state_run[state_id]["result"])

        state_id = "test_|-b_|-b_|-nop"
        self.assertIn(state_id, state_run)
        self.assertEqual(
            state_run[state_id]["comment"],
            "One or more requisite failed: issue-46762.c",
        )
        self.assertFalse(state_run[state_id]["result"])

        state_id = "test_|-c_|-c_|-nop"
        self.assertIn(state_id, state_run)
        self.assertEqual(
            state_run[state_id]["comment"],
            "One or more requisite failed: issue-46762.a",
        )
        self.assertFalse(state_run[state_id]["result"])

    @slowTest
    def test_state_nonbase_environment(self):
        """
        test state.sls with saltenv using a nonbase environment
        with a salt source
        """
        filename = os.path.join(RUNTIME_VARS.TMP, "nonbase_env")
        try:
            ret = self.run_function("state.sls", mods="non-base-env", saltenv="prod")
            ret = ret[next(iter(ret))]
            assert ret["result"]
            assert ret["comment"] == "File {0} updated".format(filename)
            assert os.path.isfile(filename)
        finally:
            try:
                os.remove(filename)
            except OSError:
                pass

    @skipIf(
        sys.platform.startswith("win"),
        "Skipped until parallel states can be fixed on Windows",
    )
    @skipIf(
        salt.utils.platform.is_darwin() and six.PY2, "This test hangs on OS X on Py2"
    )
    @slowTest
    def test_parallel_state_with_long_tag(self):
        """
        This tests the case where the state being executed has a long ID dec or
        name and states are being run in parallel. The filenames used for the
        parallel state cache were previously based on the tag for each chunk,
        and longer ID decs or name params can cause the cache file to be longer
        than the operating system's max file name length. To counter this we
        instead generate a SHA1 hash of the chunk's tag to use as the cache
        filename. This test will ensure that long tags don't cause caching
        failures.

        See https://github.com/saltstack/salt/issues/49738 for more info.
        """
        short_command = "helloworld"
        long_command = short_command * 25

        ret = self.run_function(
            "state.sls",
            mods="issue-49738",
            pillar={"short_command": short_command, "long_command": long_command},
        )
        comments = sorted([x["comment"] for x in six.itervalues(ret)])
        expected = sorted(
            ['Command "{0}" run'.format(x) for x in (short_command, long_command)]
        )
        assert comments == expected, "{0} != {1}".format(comments, expected)

    def _add_runtime_pillar(self, pillar):
        """
        helper class to add pillar data at runtime
        """
        import salt.utils.yaml

        with salt.utils.files.fopen(
            os.path.join(RUNTIME_VARS.TMP_PILLAR_TREE, "pillar.sls"), "w"
        ) as fp:
            salt.utils.yaml.safe_dump(pillar, fp)

        with salt.utils.files.fopen(
            os.path.join(RUNTIME_VARS.TMP_PILLAR_TREE, "top.sls"), "w"
        ) as fp:
            fp.write(
                textwrap.dedent(
                    """\
                     base:
                       '*':
                         - pillar
                     """
                )
            )

        self.run_function("saltutil.refresh_pillar")
        self.run_function("test.sleep", [5])

    @slowTest
    def test_state_sls_id_test(self):
        """
        test state.sls_id when test is set
        to true in pillar data
        """
        self._add_runtime_pillar(pillar={"test": True})
        testfile = os.path.join(RUNTIME_VARS.TMP, "testfile")
        comment = "The file {0} is set to be changed\nNote: No changes made, actual changes may\nbe different due to other states.".format(
            testfile
        )
        ret = self.run_function("state.sls", ["core"])

        for key, val in ret.items():
            self.assertEqual(val["comment"], comment)
            self.assertEqual(val["changes"], {"newfile": testfile})

    @slowTest
    def test_state_sls_id_test_state_test_post_run(self):
        """
        test state.sls_id when test is set to
        true post the state already being run previously
        """
        file_name = os.path.join(RUNTIME_VARS.TMP, "testfile")
        ret = self.run_function("state.sls", ["core"])
        for key, val in ret.items():
            self.assertEqual(val["comment"], "File {0} updated".format(file_name))
            self.assertEqual(val["changes"]["diff"], "New file")

        self._add_runtime_pillar(pillar={"test": True})
        ret = self.run_function("state.sls", ["core"])

        for key, val in ret.items():
            self.assertEqual(
                val["comment"], "The file {0} is in the correct state".format(file_name)
            )
            self.assertEqual(val["changes"], {})

    @slowTest
    def test_state_sls_id_test_true(self):
        """
        test state.sls_id when test=True is passed as arg
        """
        file_name = os.path.join(RUNTIME_VARS.TMP, "testfile")
        ret = self.run_function("state.sls", ["core"], test=True)
        for key, val in ret.items():
            self.assertEqual(
                val["comment"],
                "The file {0} is set to be changed\nNote: No changes made, actual changes may\nbe different due to other states.".format(
                    file_name
                ),
            )
            self.assertEqual(val["changes"], {"newfile": file_name})

    @slowTest
    def test_state_sls_id_test_true_post_run(self):
        """
        test state.sls_id when test is set to true as an
        arg post the state already being run previously
        """
        file_name = os.path.join(RUNTIME_VARS.TMP, "testfile")
        ret = self.run_function("state.sls", ["core"])
        for key, val in ret.items():
            self.assertEqual(val["comment"], "File {0} updated".format(file_name))
            self.assertEqual(val["changes"]["diff"], "New file")

        ret = self.run_function("state.sls", ["core"], test=True)

        for key, val in ret.items():
            self.assertEqual(
                val["comment"], "The file {0} is in the correct state".format(file_name)
            )
            self.assertEqual(val["changes"], {})

    @slowTest
    def test_state_sls_id_test_false_pillar_true(self):
        """
        test state.sls_id when test is set to false as an
        arg and minion_state_test is set to True. Should
        return test=False.
        """
        file_name = os.path.join(RUNTIME_VARS.TMP, "testfile")
        self._add_runtime_pillar(pillar={"test": True})
        ret = self.run_function("state.sls", ["core"], test=False)

        for key, val in ret.items():
            self.assertEqual(val["comment"], "File {0} updated".format(file_name))
            self.assertEqual(val["changes"]["diff"], "New file")

    def test_state_test_pillar_false(self):
        """
        test state.test forces test kwarg to True even when pillar is set to False
        """
        self._add_runtime_pillar(pillar={"test": False})
        testfile = os.path.join(RUNTIME_VARS.TMP, "testfile")
        comment = "The file {0} is set to be changed\nNote: No changes made, actual changes may\nbe different due to other states.".format(
            testfile
        )
        ret = self.run_function("state.test", ["core"])
        self.assertIsInstance(ret, dict)

        for key, val in ret.items():
            self.assertEqual(val["comment"], comment)
            self.assertEqual(val["changes"], {"newfile": testfile})

    def test_state_test_test_false_pillar_false(self):
        """
        test state.test forces test kwarg to True even when pillar and kwarg are set
        to False
        """
        self._add_runtime_pillar(pillar={"test": False})
        testfile = os.path.join(RUNTIME_VARS.TMP, "testfile")
        comment = "The file {0} is set to be changed\nNote: No changes made, actual changes may\nbe different due to other states.".format(
            testfile
        )
        ret = self.run_function("state.test", ["core"], test=False)

        for key, val in ret.items():
            self.assertEqual(val["comment"], comment)
            self.assertEqual(val["changes"], {"newfile": testfile})

    @skipIf(
        six.PY3 and salt.utils.platform.is_darwin(), "Test is broken on macosx and PY3"
    )
    @slowTest
    def test_issue_30161_unless_and_onlyif_together(self):
        """
        test cmd.run using multiple unless options where the first cmd in the
        list will pass, but the second will fail. This tests the fix for issue
        #35384. (The fix is in PR #35545.)
        """
        sls = self.run_function("state.sls", mods="issue-30161")
        self.assertSaltTrueReturn(sls)
        # We must assert against the comment here to make sure the comment reads that the
        # command "echo "hello"" was run. This ensures that we made it to the last unless
        # command in the state. If the comment reads "unless condition is true", or similar,
        # then the unless state run bailed out after the first unless command succeeded,
        # which is the bug we're regression testing for.
        _expected = {
            "file_|-unless_false_onlyif_false_|-{0}{1}test.txt_|-managed".format(
                RUNTIME_VARS.TMP, os.path.sep
            ): {
                "comment": "onlyif condition is false\nunless condition is false",
                "name": "{0}{1}test.txt".format(RUNTIME_VARS.TMP, os.path.sep),
                "skip_watch": True,
                "changes": {},
                "result": True,
            },
            "file_|-unless_false_onlyif_true_|-{0}{1}test.txt_|-managed".format(
                RUNTIME_VARS.TMP, os.path.sep
            ): {
                "comment": "Empty file",
                "name": "{0}{1}test.txt".format(RUNTIME_VARS.TMP, os.path.sep),
                "start_time": "18:10:20.341753",
                "result": True,
                "changes": {
                    "new": "file {0}{1}test.txt created".format(
                        RUNTIME_VARS.TMP, os.path.sep
                    )
                },
            },
            "file_|-unless_true_onlyif_false_|-{0}{1}test.txt_|-managed".format(
                RUNTIME_VARS.TMP, os.path.sep
            ): {
                "comment": "onlyif condition is false\nunless condition is true",
                "name": "{0}{1}test.txt".format(RUNTIME_VARS.TMP, os.path.sep),
                "start_time": "18:10:22.936446",
                "skip_watch": True,
                "changes": {},
                "result": True,
            },
            "file_|-unless_true_onlyif_true_|-{0}{1}test.txt_|-managed".format(
                RUNTIME_VARS.TMP, os.path.sep
            ): {
                "comment": "onlyif condition is true\nunless condition is true",
                "name": "{0}{1}test.txt".format(RUNTIME_VARS.TMP, os.path.sep),
                "skip_watch": True,
                "changes": {},
                "result": True,
            },
        }
        for id in _expected:
            self.assertEqual(sls[id]["comment"], _expected[id]["comment"])

    @skipIf(
        six.PY3 and salt.utils.platform.is_darwin(), "Test is broken on macosx and PY3"
    )
    @slowTest
    def test_state_sls_unicode_characters(self):
        """
        test state.sls when state file contains non-ascii characters
        """
        ret = self.run_function("state.sls", ["issue-46672"])
        log.debug("== ret %s ==", type(ret))

        _expected = "cmd_|-echo1_|-echo 'This is Æ test!'_|-run"
        self.assertIn(_expected, ret)

    @skipIf(
        six.PY3 and salt.utils.platform.is_darwin(), "Test is broken on macosx and PY3"
    )
    @slowTest
    def test_state_sls_unicode_characters_cmd_output(self):
        """
        test the output from running and echo command with non-ascii
        characters.
        """
        ret = self.run_function("state.sls", ["issue-46672-a"], timeout=60)
        key = list(ret.keys())[0]
        log.debug("== ret %s ==", type(ret))
        _expected = "This is Æ test!"
        if salt.utils.platform.is_windows():
            # Windows cmd.exe will mangle the output using cmd's codepage.
            if six.PY2:
                _expected = "'This is A+ test!'"
            else:
                _expected = "'This is ’ test!'"
        self.assertEqual(_expected, ret[key]["changes"]["stdout"])

    def tearDown(self):
        rm_files = [
            os.path.join(RUNTIME_VARS.TMP, "nonbase_env"),
            os.path.join(RUNTIME_VARS.TMP, "testfile"),
            os.path.join(RUNTIME_VARS.TMP, "test.txt"),
            os.path.join(RUNTIME_VARS.TMP_STATE_TREE, "top.sls"),
        ]

        for file_ in rm_files:
            if os.path.isfile(file_):
                os.remove(file_)

        # remove old pillar data
        for filename in os.listdir(RUNTIME_VARS.TMP_PILLAR_TREE):
            os.remove(os.path.join(RUNTIME_VARS.TMP_PILLAR_TREE, filename))
        self.run_function("saltutil.refresh_pillar")
        self.run_function("test.sleep", [5])

    @slowTest
    def test_state_sls_integer_name(self):
        """
        This tests the case where the state file is named
        only with integers
        """
        state_run = self.run_function("state.sls", mods="12345")

        state_id = "test_|-always-passes_|-always-passes_|-succeed_without_changes"
        self.assertIn(state_id, state_run)
        self.assertEqual(state_run[state_id]["comment"], "Success!")
        self.assertTrue(state_run[state_id]["result"])

    @slowTest
    def test_state_sls_lazyloader_allows_recursion(self):
        """
        This tests that referencing dunders like __salt__ work
        context: https://github.com/saltstack/salt/pull/51499
        """
        state_run = self.run_function("state.sls", mods="issue-51499")

        state_id = "test_|-always-passes_|-foo_|-succeed_without_changes"
        self.assertIn(state_id, state_run)
        self.assertEqual(state_run[state_id]["comment"], "Success!")
        self.assertTrue(state_run[state_id]["result"])

    @slowTest
    def test_issue_56131(self):
        module_path = os.path.join(RUNTIME_VARS.CODE_DIR, "pip.py")
        if six.PY3:
            modulec_path = os.path.join(RUNTIME_VARS.CODE_DIR, "__pycache__", "pip.pyc")
        else:
            modulec_path = os.path.join(RUNTIME_VARS.CODE_DIR, "pip.pyc")
        unzip_path = os.path.join(RUNTIME_VARS.TMP, "issue-56131.txt")

        def clean_paths(paths):
            for path in paths:
                try:
                    os.remove(path)
                except OSError:
                    log.warn("Path not found: %s", path)

        with salt.utils.files.fopen(module_path, "w") as fp:
            fp.write('raise ImportError("No module named pip")')
        self.addCleanup(clean_paths, [unzip_path, module_path, modulec_path])
        assert not os.path.exists(unzip_path)
        state_run = self.run_function(
            "state.sls",
            mods="issue-56131",
            pillar={"unzip_to": RUNTIME_VARS.TMP},
            timeout=30,
        )
        assert state_run is not False
        assert os.path.exists(unzip_path)

    @slowTest
    def test_jinja_renderer_argline(self):
        """
        This is a test case for https://github.com/saltstack/salt/issues/55124

        Renderer for this is in tests/integration/files/file/base/_renderers/issue55124.py
        """
        result = self.run_function("state.sls", mods="issue-55124")
        assert isinstance(result, dict), result
        result = result[next(iter(result))]
        assert result["result"], result



# coding: utf-8

# Python libs
from __future__ import absolute_import

# Salt libs
import salt.beacons.adb as adb
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import Mock, patch

# Salt testing libs
from tests.support.unit import TestCase


class ADBBeaconTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test case for salt.beacons.adb
    """

    def setup_loader_modules(self):
        return {adb: {"last_state": {}, "last_state_extra": {"no_devices": False}}}

    def test_no_adb_command(self):
        with patch("salt.utils.path.which") as mock:
            mock.return_value = None

            ret = adb.__virtual__()

            mock.assert_called_once_with("adb")
            self.assertFalse(ret)

    def test_with_adb_command(self):
        with patch("salt.utils.path.which") as mock:
            mock.return_value = "/usr/bin/adb"

            ret = adb.__virtual__()

            mock.assert_called_once_with("adb")
            self.assertEqual(ret, "adb")

    def test_non_list_config(self):
        config = {}

        ret = adb.validate(config)

        self.assertEqual(ret, (False, "Configuration for adb beacon must be a list."))

    def test_empty_config(self):
        config = [{}]

        ret = adb.validate(config)

        self.assertEqual(
            ret, (False, "Configuration for adb beacon must include a states array.")
        )

    def test_invalid_states(self):
        config = [{"states": ["Random", "Failings"]}]

        ret = adb.validate(config)

        self.assertEqual(
            ret,
            (
                False,
                "Need a one of the following"
                " adb states: offline, bootloader,"
                " device, host, recovery, no"
                " permissions, sideload,"
                " unauthorized, unknown, missing",
            ),
        )

    def test_device_state(self):
        config = [{"states": ["device"]}]

        mock = Mock(return_value="List of devices attached\nHTC\tdevice",)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):

            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

    def test_device_state_change(self):
        config = [{"states": ["offline"]}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "List of devices attached\nHTC\toffline",
        ]

        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):

            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "offline", "tag": "offline"}]
            )

    def test_multiple_devices(self):
        config = [{"states": ["offline", "device"]}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "List of devices attached\nHTC\toffline\nNexus\tdevice",
        ]

        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):

            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(
                ret,
                [
                    {"device": "HTC", "state": "offline", "tag": "offline"},
                    {"device": "Nexus", "state": "device", "tag": "device"},
                ],
            )

    def test_no_devices_with_different_states(self):
        config = [{"states": ["offline"], "no_devices_event": True}]

        mock = Mock(return_value="List of devices attached\nHTC\tdevice")
        with patch.dict(adb.__salt__, {"cmd.run": mock}):

            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_no_devices_no_repeat(self):
        config = [{"states": ["offline", "device"], "no_devices_event": True}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "List of devices attached",
            "List of devices attached",
        ]

        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):

            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(ret, [{"tag": "no_devices"}])

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_no_devices(self):
        config = [{"states": ["offline", "device"], "no_devices_event": True}]

        out = ["List of devices attached", "List of devices attached"]

        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):

            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(ret, [{"tag": "no_devices"}])

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_device_missing(self):
        config = [{"states": ["device", "missing"]}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "List of devices attached",
            "List of devices attached\nHTC\tdevice",
            "List of devices attached\nHTC\tdevice",
        ]

        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):

            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "missing", "tag": "missing"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_with_startup(self):
        config = [{"states": ["device"]}]

        mock = Mock(
            return_value="* daemon started successfully *\nList of devices attached\nHTC\tdevice",
        )
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

    def test_with_user(self):
        config = [{"states": ["device"], "user": "fred"}]

        mock = Mock(
            return_value="* daemon started successfully *\nList of devices attached\nHTC\tdevice"
        )
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            mock.assert_called_once_with("adb devices", runas="fred")
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

    def test_device_low_battery(self):
        config = [{"states": ["device"], "battery_low": 30}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret,
                [
                    {"device": "HTC", "state": "device", "tag": "device"},
                    {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
                ],
            )

    def test_device_no_repeat(self):
        config = [{"states": ["device"], "battery_low": 30}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25",
            "List of devices attached\nHTC\tdevice",
            "25",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret,
                [
                    {"device": "HTC", "state": "device", "tag": "device"},
                    {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
                ],
            )

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_device_no_repeat_capacity_increase(self):
        config = [{"states": ["device"], "battery_low": 75}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25",
            "List of devices attached\nHTC\tdevice",
            "30",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret,
                [
                    {"device": "HTC", "state": "device", "tag": "device"},
                    {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
                ],
            )

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_device_no_repeat_with_not_found_state(self):
        config = [{"states": ["offline"], "battery_low": 30}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25",
            "List of devices attached\nHTC\tdevice",
            "25",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "battery_level": 25, "tag": "battery_low"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_device_battery_charged(self):
        config = [{"states": ["device"], "battery_low": 30}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "100",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

    def test_device_low_battery_equal(self):
        config = [{"states": ["device"], "battery_low": 25}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret,
                [
                    {"device": "HTC", "state": "device", "tag": "device"},
                    {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
                ],
            )

    def test_device_battery_not_found(self):
        config = [{"states": ["device"], "battery_low": 25}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "/system/bin/sh: cat: /sys/class/power_supply/*/capacity: No such file or directory",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

    def test_device_repeat_multi(self):
        config = [{"states": ["offline"], "battery_low": 35}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25",
            "List of devices attached\nHTC\tdevice",
            "40",
            "List of devices attached\nHTC\tdevice",
            "25",
            "List of devices attached\nHTC\tdevice",
            "80",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "battery_level": 25, "tag": "battery_low"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "battery_level": 25, "tag": "battery_low"}]
            )

            ret = adb.beacon(config)
            self.assertEqual(ret, [])

    def test_weird_batteries(self):
        config = [{"states": ["device"], "battery_low": 25}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "-9000",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret, [{"device": "HTC", "state": "device", "tag": "device"}]
            )

    def test_multiple_batteries(self):
        config = [{"states": ["device"], "battery_low": 30}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25\n40",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret,
                [
                    {"device": "HTC", "state": "device", "tag": "device"},
                    {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
                ],
            )

    def test_multiple_low_batteries(self):
        config = [{"states": ["device"], "battery_low": 30}]

        out = [
            "List of devices attached\nHTC\tdevice",
            "25\n14",
        ]
        mock = Mock(side_effect=out)
        with patch.dict(adb.__salt__, {"cmd.run": mock}):
            ret = adb.validate(config)
            self.assertEqual(ret, (True, "Valid beacon configuration"))

            ret = adb.beacon(config)
            self.assertEqual(
                ret,
                [
                    {"device": "HTC", "state": "device", "tag": "device"},
                    {"device": "HTC", "battery_level": 25, "tag": "battery_low"},
                ],
            )



# -*- coding: utf-8 -*-
"""
Manage Apache Sites

.. versionadded:: 2016.3.0

Enable and disable apache sites.

.. code-block:: yaml

    Enable default site:
      apache_site.enabled:
        - name: default

    Disable default site:
      apache_site.disabled:
        - name: default
"""
from __future__ import absolute_import, print_function, unicode_literals

# Import salt libs
from salt.ext import six


def __virtual__():
    """
    Only load if a2ensite is available.
    """
    if "apache.a2ensite" in __salt__:
        return "apache_site"
    return (False, "apache module could not be loaded")


def enabled(name):
    """
    Ensure an Apache site is enabled.

    name
        Name of the Apache site
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_enabled = __salt__["apache.check_site_enabled"](name)
    if not is_enabled:
        if __opts__["test"]:
            msg = "Apache site {0} is set to be enabled.".format(name)
            ret["comment"] = msg
            ret["changes"]["old"] = None
            ret["changes"]["new"] = name
            ret["result"] = None
            return ret
        status = __salt__["apache.a2ensite"](name)["Status"]
        if isinstance(status, six.string_types) and "enabled" in status:
            ret["result"] = True
            ret["changes"]["old"] = None
            ret["changes"]["new"] = name
        else:
            ret["result"] = False
            ret["comment"] = "Failed to enable {0} Apache site".format(name)
            if isinstance(status, six.string_types):
                ret["comment"] = ret["comment"] + " ({0})".format(status)
            return ret
    else:
        ret["comment"] = "{0} already enabled.".format(name)
    return ret


def disabled(name):
    """
    Ensure an Apache site is disabled.

    name
        Name of the Apache site
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_enabled = __salt__["apache.check_site_enabled"](name)
    if is_enabled:
        if __opts__["test"]:
            msg = "Apache site {0} is set to be disabled.".format(name)
            ret["comment"] = msg
            ret["changes"]["old"] = name
            ret["changes"]["new"] = None
            ret["result"] = None
            return ret
        status = __salt__["apache.a2dissite"](name)["Status"]
        if isinstance(status, six.string_types) and "disabled" in status:
            ret["result"] = True
            ret["changes"]["old"] = name
            ret["changes"]["new"] = None
        else:
            ret["result"] = False
            ret["comment"] = "Failed to disable {0} Apache site".format(name)
            if isinstance(status, six.string_types):
                ret["comment"] = ret["comment"] + " ({0})".format(status)
            return ret
    else:
        ret["comment"] = "{0} already disabled.".format(name)
    return ret



# -*- coding: utf-8 -*-
"""
    :synopsis: Unit Tests for Windows PKI Module 'state.win_pki'
    :platform: Windows
    :maturity: develop
    .. versionadded:: 2017.7.0
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.win_pki as win_pki

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

CERT_PATH = r"C:\certs\testdomain.local.cer"
THUMBPRINT = "9988776655443322111000AAABBBCCCDDDEEEFFF"
STORE_PATH = r"Cert:\LocalMachine\My"

CERTS = {
    THUMBPRINT: {
        "dnsnames": ["testdomain.local"],
        "serialnumber": "0123456789AABBCCDD",
        "subject": "CN=testdomain.local, OU=testou, O=testorg, S=California, C=US",
        "thumbprint": THUMBPRINT,
        "version": 3,
    }
}


class WinPkiTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.win_pki
    """

    def setup_loader_modules(self):
        return {win_pki: {}}

    def test_import_cert(self):
        """
        Test - Import the certificate file into the given certificate store.
        """
        kwargs = {"name": CERT_PATH}
        ret = {
            "name": kwargs["name"],
            "changes": {"old": None, "new": THUMBPRINT},
            "comment": "Certificate '{0}' imported into store: {1}".format(
                THUMBPRINT, STORE_PATH
            ),
            "result": True,
        }
        mock_cache_file = MagicMock(return_value=CERT_PATH)
        mock_certs = MagicMock(return_value={})
        mock_cert_file = MagicMock(return_value=CERTS[THUMBPRINT])
        mock_import_cert = MagicMock(return_value=True)
        with patch.dict(
            win_pki.__salt__,
            {
                "cp.cache_file": mock_cache_file,
                "win_pki.get_certs": mock_certs,
                "win_pki.get_cert_file": mock_cert_file,
                "win_pki.import_cert": mock_import_cert,
            },
        ):
            with patch.dict(win_pki.__opts__, {"test": False}):
                self.assertEqual(win_pki.import_cert(**kwargs), ret)
            with patch.dict(win_pki.__opts__, {"test": True}):
                ret["comment"] = (
                    "Certificate '{0}' will be imported into store:" " {1}"
                ).format(THUMBPRINT, STORE_PATH)
                ret["result"] = None
                self.assertEqual(win_pki.import_cert(**kwargs), ret)

    def test_remove_cert(self):
        """
        Test - Remove the certificate from the given certificate store.
        """
        kwargs = {"name": "remove-cert", "thumbprint": THUMBPRINT}
        ret = {
            "name": kwargs["name"],
            "changes": {"old": kwargs["thumbprint"], "new": None},
            "comment": "Certificate '{0}' removed from store: {1}".format(
                kwargs["thumbprint"], STORE_PATH
            ),
            "result": True,
        }
        mock_certs = MagicMock(return_value=CERTS)
        mock_remove_cert = MagicMock(return_value=True)
        with patch.dict(
            win_pki.__salt__,
            {"win_pki.get_certs": mock_certs, "win_pki.remove_cert": mock_remove_cert},
        ):
            with patch.dict(win_pki.__opts__, {"test": False}):
                self.assertEqual(win_pki.remove_cert(**kwargs), ret)
            with patch.dict(win_pki.__opts__, {"test": True}):
                ret["comment"] = (
                    "Certificate '{0}' will be removed from store:" " {1}"
                ).format(kwargs["thumbprint"], STORE_PATH)
                ret["result"] = None
                self.assertEqual(win_pki.remove_cert(**kwargs), ret)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    runtests_helpers.py
    ~~~~~~~~~~~~~~~~~~~
"""

# Import python libs
from __future__ import absolute_import

import fnmatch
import logging
import os
import re
import sys
import tempfile

# Import salt libs
import salt.utils.platform

# Import 3rd-party libs
from salt.ext import six

# Import tests libs
try:
    from tests.support.runtests import RUNTIME_VARS
except ImportError:
    # Salt SSH Tests
    SYS_TMP_DIR = os.path.realpath(
        # Avoid ${TMPDIR} and gettempdir() on MacOS as they yield a base path too long
        # for unix sockets: ``error: AF_UNIX path too long``
        # Gentoo Portage prefers ebuild tests are rooted in ${TMPDIR}
        os.environ.get("TMPDIR", tempfile.gettempdir())
        if not salt.utils.platform.is_darwin()
        else "/tmp"
    )
    # This tempdir path is defined on tests.integration.__init__
    TMP = os.path.join(SYS_TMP_DIR, "salt-tests-tmpdir")

    class RUNTIME_VARS(object):
        TMP = TMP
        SYS_TMP_DIR = SYS_TMP_DIR


log = logging.getLogger(__name__)


def get_salt_temp_dir():
    return RUNTIME_VARS.TMP


def get_salt_temp_dir_for_path(*path):
    return os.path.join(RUNTIME_VARS.TMP, *path)


def get_sys_temp_dir_for_path(*path):
    return os.path.join(RUNTIME_VARS.SYS_TMP_DIR, *path)


def get_invalid_docs():
    """
    Outputs the functions which do not have valid CLI example, or are missing a
    docstring.
    """
    allow_failure = (
        "cmd.win_runas",
        "cp.recv",
        "cp.recv_chunked",
        "glance.warn_until",
        "ipset.long_range",
        "libcloud_compute.get_driver",
        "libcloud_dns.get_driver",
        "libcloud_loadbalancer.get_driver",
        "libcloud_storage.get_driver",
        "log.critical",
        "log.debug",
        "log.error",
        "log.exception",
        "log.info",
        "log.warning",
        "lowpkg.bin_pkg_info",
        "lxc.run_cmd",
        "mantest.install",
        "mantest.search",
        "nspawn.restart",
        "nspawn.stop",
        "pkg.expand_repo_def",
        "pip.iteritems",
        "pip.parse_version",
        "peeringdb.clean_kwargs",
        "runtests_decorators.depends",
        "runtests_decorators.depends_will_fallback",
        "runtests_decorators.missing_depends",
        "runtests_decorators.missing_depends_will_fallback",
        "state.apply",
        "status.list2cmdline",
        "swift.head",
        "test.rand_str",
        "travisci.parse_qs",
        "vsphere.clean_kwargs",
        "vsphere.disconnect",
        "vsphere.get_service_instance_via_proxy",
        "vsphere.gets_service_instance_via_proxy",
        "vsphere.supports_proxies",
        "vsphere.test_vcenter_connection",
        "vsphere.wraps",
    )
    allow_failure_glob = (
        "runtests_decorators.*",
        "runtests_helpers.*",
        "vsphere.*",
    )
    nodoc = set()
    noexample = set()
    for fun, docstring in six.iteritems(__salt__["sys.doc"]()):
        if fun in allow_failure:
            continue
        else:
            for pat in allow_failure_glob:
                if fnmatch.fnmatch(fun, pat):
                    matched_glob = True
                    break
            else:
                matched_glob = False
            if matched_glob:
                continue
        if not isinstance(docstring, six.string_types):
            nodoc.add(fun)
        elif isinstance(docstring, dict) and not re.search(
            r"([E|e]xample(?:s)?)+(?:.*):?", docstring
        ):
            noexample.add(fun)

    return {
        "missing_docstring": sorted(nodoc),
        "missing_cli_example": sorted(noexample),
    }


def modules_available(*names):
    """
    Returns a list of modules not available. Empty list if modules are all available
    """
    not_found = []
    for name in names:
        if "." not in name:
            name = name + ".*"
        if not fnmatch.filter(list(__salt__), name):
            not_found.append(name)
    return not_found


def nonzero_retcode_return_true():
    """
    Sets a nonzero retcode before returning. Designed to test orchestration.
    """
    __context__["retcode"] = 1
    return True


def nonzero_retcode_return_false():
    """
    Sets a nonzero retcode before returning. Designed to test orchestration.
    """
    __context__["retcode"] = 1
    return False


def fail_function(*args, **kwargs):  # pylint: disable=unused-argument
    """
    Return False no matter what is passed to it
    """
    return False


def get_python_executable():
    """
    Return the path to the python executable.

    This is particularly important when running the test suite within a virtualenv, while trying
    to create virtualenvs on windows.
    """
    try:
        if salt.utils.platform.is_windows():
            python_binary = os.path.join(
                sys.real_prefix, os.path.basename(sys.executable)
            )
        else:
            python_binary = os.path.join(
                sys.real_prefix, "bin", os.path.basename(sys.executable)
            )
            if not os.path.exists(python_binary):
                if not python_binary[-1].isdigit():
                    versioned_python_binary = "{}{}".format(
                        python_binary, *sys.version_info
                    )
                    log.info(
                        "Python binary could not be found at %s. Trying %s",
                        python_binary,
                        versioned_python_binary,
                    )
                    if os.path.exists(versioned_python_binary):
                        python_binary = versioned_python_binary
        if not os.path.exists(python_binary):
            log.warning("Python binary could not be found at %s", python_binary)
            python_binary = None
    except AttributeError:
        # We're not running inside a virtualenv
        python_binary = sys.executable
    return python_binary



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.rabbitmq_plugin as rabbitmq_plugin

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class RabbitmqPluginTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.rabbitmq_plugin
    """

    def setup_loader_modules(self):
        return {rabbitmq_plugin: {}}

    # 'enabled' function tests: 1

    def test_enabled(self):
        """
        Test to ensure the RabbitMQ plugin is enabled.
        """
        name = "some_plugin"

        ret = {"name": name, "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(side_effect=[True, False])
        with patch.dict(rabbitmq_plugin.__salt__, {"rabbitmq.plugin_is_enabled": mock}):
            comment = "Plugin 'some_plugin' is already enabled."
            ret.update({"comment": comment})
            self.assertDictEqual(rabbitmq_plugin.enabled(name), ret)

            with patch.dict(rabbitmq_plugin.__opts__, {"test": True}):
                comment = "Plugin 'some_plugin' is set to be enabled."
                changes = {"new": "some_plugin", "old": ""}
                ret.update({"comment": comment, "result": None, "changes": changes})
                self.assertDictEqual(rabbitmq_plugin.enabled(name), ret)

    # 'disabled' function tests: 1

    def test_disabled(self):
        """
        Test to ensure the RabbitMQ plugin is disabled.
        """
        name = "some_plugin"

        ret = {"name": name, "changes": {}, "result": True, "comment": ""}

        mock = MagicMock(side_effect=[False, True])
        with patch.dict(rabbitmq_plugin.__salt__, {"rabbitmq.plugin_is_enabled": mock}):
            comment = "Plugin 'some_plugin' is already disabled."
            ret.update({"comment": comment})
            self.assertDictEqual(rabbitmq_plugin.disabled(name), ret)

            with patch.dict(rabbitmq_plugin.__opts__, {"test": True}):
                comment = "Plugin 'some_plugin' is set to be disabled."
                changes = {"new": "", "old": "some_plugin"}
                ret.update({"comment": comment, "result": None, "changes": changes})
                self.assertDictEqual(rabbitmq_plugin.disabled(name), ret)



# -*- coding: utf-8 -*-
"""
Module for running arbitrary tests with a __virtual__ function
"""
from __future__ import absolute_import, print_function, unicode_literals


def __virtual__():
    return (False, "The test_virtual execution module failed to load.")


def ping():
    return True



# -*- coding: utf-8 -*-
"""
    :codeauthor: Ryan Lewis (ryansname@gmail.com)

    tests.unit.modules.portage_flags
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import re

# Import salt libs
import salt.modules.portage_config as portage_config
import salt.utils.files
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch

# Import Salt Testing libs
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class PortageConfigTestCase(TestCase, LoaderModuleMockMixin):
    class DummyAtom(object):
        def __init__(self):
            self.cp = None
            self.repo = None

        def __call__(self, atom, *_, **__):
            if atom == "#" or isinstance(atom, MagicMock):
                self.repo = None
                self.cp = None
                return self

            # extract (and remove) repo
            atom, self.repo = atom.split("::") if "::" in atom else (atom, None)

            # remove '>, >=, <=, =, ~' etc.
            atom = re.sub(r"[<>~+=]", "", atom)
            # remove slots
            atom = re.sub(r":[0-9][^:]*", "", atom)
            # remove version
            atom = re.sub(r"-[0-9][\.0-9]*", "", atom)

            self.cp = atom
            return self

    def setup_loader_modules(self):
        try:
            import portage  # pylint: disable=unused-import

            return {}
        except ImportError:
            dummy_atom = self.DummyAtom()
            self.portage = MagicMock()
            self.portage.dep.Atom = MagicMock(side_effect=dummy_atom)
            self.portage.dep_getkey = MagicMock(side_effect=lambda x: dummy_atom(x).cp)
            self.portage.exception.InvalidAtom = Exception
            self.addCleanup(delattr, self, "portage")
            return {portage_config: {"portage": self.portage}}

    def test_get_config_file_wildcards(self):
        pairs = [
            ("*/*::repo", "/etc/portage/package.mask/repo"),
            ("*/pkg::repo", "/etc/portage/package.mask/pkg"),
            ("cat/*", "/etc/portage/package.mask/cat_"),
            ("cat/pkg", "/etc/portage/package.mask/cat/pkg"),
            ("cat/pkg::repo", "/etc/portage/package.mask/cat/pkg"),
        ]

        for (atom, expected) in pairs:
            self.assertEqual(portage_config._get_config_file("mask", atom), expected)

    def test_enforce_nice_config(self):
        atoms = [
            ("*/*::repo", "repo"),
            ("*/pkg1::repo", "pkg1"),
            ("cat/*", "cat_"),
            ("cat/pkg2", "cat/pkg2"),
            ("cat/pkg3::repo", "cat/pkg3"),
            ("<cat/pkg4-0.0.0.0", "cat/pkg4"),
            (">cat/pkg5-0.0.0.0:0", "cat/pkg5"),
            (">cat/pkg6-0.0.0.0:0::repo", "cat/pkg6"),
            ("<=cat/pkg7-0.0.0.0", "cat/pkg7"),
            ("=cat/pkg8-0.0.0.0", "cat/pkg8"),
        ]

        supported = [
            ("accept_keywords", ["~amd64"]),
            ("env", ["glibc.conf"]),
            ("license", ["LICENCE1", "LICENCE2"]),
            ("mask", [""]),
            ("properties", ["* -interactive"]),
            ("unmask", [""]),
            ("use", ["apple", "-banana", "ananas", "orange"]),
        ]

        base_path = RUNTIME_VARS.TMP + "/package.{0}"

        def make_line(atom, addition):
            return atom + (" " + addition if addition != "" else "") + "\n"

        for typ, additions in supported:
            path = base_path.format(typ)
            with salt.utils.files.fopen(path, "a") as fh:
                for atom, _ in atoms:
                    for addition in additions:
                        line = make_line(atom, addition)
                        fh.write("# comment for: " + line)
                        fh.write(line)

        with patch.object(portage_config, "BASE_PATH", base_path):
            with patch.object(
                portage_config, "_merge_flags", lambda l1, l2, _: list(set(l1 + l2))
            ):
                portage_config.enforce_nice_config()

        for typ, additions in supported:
            for atom, file_name in atoms:
                with salt.utils.files.fopen(
                    base_path.format(typ) + "/" + file_name, "r"
                ) as fh:
                    for line in fh:
                        self.assertTrue(
                            atom in line, msg="'{}' not in '{}'".format(addition, line)
                        )
                        for addition in additions:
                            self.assertTrue(
                                addition in line,
                                msg="'{}' not in '{}'".format(addition, line),
                            )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.modules.debian_service as debian_service

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class DebianServicesTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.debian_service
    """

    def setup_loader_modules(self):
        return {debian_service: {}}

    def test_get_enabled(self):
        """
        Test for Return a list of service that are enabled on boot
        """
        mock_runlevel = MagicMock(return_value=1)
        mock_prefix = "/etc/rc1.d/S"
        mock_glob = MagicMock(return_value=[mock_prefix + "01name"])

        with patch.object(debian_service, "_get_runlevel", mock_runlevel):
            with patch.object(debian_service.glob, "glob", mock_glob):
                self.assertEqual(debian_service.get_enabled()[0], "name")

    def test_get_disabled(self):
        """
        Test for Return a set of services that are installed but disabled
        """
        mock = MagicMock(return_value=["A"])
        with patch.object(debian_service, "get_all", mock):
            mock = MagicMock(return_value=["B"])
            with patch.object(debian_service, "get_enabled", mock):
                self.assertEqual(debian_service.get_disabled(), ["A"])

    def test_available(self):
        """
        Test for Returns ``True`` if the specified service is
        available, otherwise returns
        ``False``.
        """
        mock = MagicMock(return_value=["A"])
        with patch.object(debian_service, "get_all", mock):
            self.assertFalse(debian_service.available("name"))

    def test_missing(self):
        """
        Test for The inverse of service.available.
        """
        mock = MagicMock(return_value=["A"])
        with patch.object(debian_service, "get_all", mock):
            self.assertTrue(debian_service.missing("name"))

    def test_getall(self):
        """
        Test for Return all available boot services
        """
        mock = MagicMock(return_value=("A"))
        with patch.object(debian_service, "get_enabled", mock):
            self.assertEqual(debian_service.get_all()[0], "A")

    def test_start(self):
        """
        Test for Start the specified service
        """
        mock = MagicMock(return_value=True)
        with patch.object(debian_service, "_service_cmd", mock):
            with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                self.assertFalse(debian_service.start("name"))

    def test_stop(self):
        """
        Test for Stop the specified service
        """
        mock = MagicMock(return_value=True)
        with patch.object(debian_service, "_service_cmd", mock):
            with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                self.assertFalse(debian_service.stop("name"))

    def test_restart(self):
        """
        Test for Restart the named service
        """
        mock = MagicMock(return_value=True)
        with patch.object(debian_service, "_service_cmd", mock):
            with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                self.assertFalse(debian_service.restart("name"))

    def test_reload_(self):
        """
        Test for Reload the named service
        """
        mock = MagicMock(return_value=True)
        with patch.object(debian_service, "_service_cmd", mock):
            with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                self.assertFalse(debian_service.reload_("name"))

    def test_force_reload(self):
        """
        Test for Force-reload the named service
        """
        mock = MagicMock(return_value=True)
        with patch.object(debian_service, "_service_cmd", mock):
            with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                self.assertFalse(debian_service.force_reload("name"))

    def test_status(self):
        """
        Test for Return the status for a service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(debian_service.__salt__, {"status.pid": mock}):
            self.assertTrue(debian_service.status("name", 1))

        mock = MagicMock(return_value="A")
        with patch.object(debian_service, "_service_cmd", mock):
            mock = MagicMock(return_value=True)
            with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                self.assertFalse(debian_service.status("name"))

    def test_enable(self):
        """
        Test for Enable the named service to start at boot
        """
        mock = MagicMock(return_value="5")
        with patch.object(debian_service, "_osrel", mock):
            mock = MagicMock(return_value="")
            with patch.object(debian_service, "_cmd_quote", mock):
                mock = MagicMock(return_value=True)
                with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                    self.assertFalse(debian_service.enable("name"))

    def test_disable(self):
        """
        Test for Disable the named service to start at boot
        """
        mock = MagicMock(return_value="5")
        with patch.object(debian_service, "_osrel", mock):
            mock = MagicMock(return_value=True)
            with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
                self.assertFalse(debian_service.disable("name"))

    def test_enabled(self):
        """
        Test for Return True if the named service is enabled, false otherwise
        """
        mock = MagicMock(return_value=["A"])
        with patch.object(debian_service, "get_enabled", mock):
            self.assertFalse(debian_service.enabled("name"))

    def test_disabled(self):
        """
        Test for Return True if the named service is enabled, false otherwise
        """
        mock = MagicMock(return_value=["A"])
        with patch.object(debian_service, "get_enabled", mock):
            self.assertFalse(debian_service.disabled("name"))
